@@ -18,7 +18,7 @@ namespace code {
 			 * Transform.
 			 */
 
-			virtual Listing *STORM_FN transform(Listing *src, MAYBE(Binary *) owner) const;
+			virtual Listing *STORM_FN transform(Listing *src) const;
 			virtual void STORM_FN output(Listing *src, Output *to) const;
 
 			/**
@@ -42,6 +42,7 @@ namespace code {
 			virtual Listing *STORM_FN engineRedirect(TypeDesc *result, Array<TypeDesc *> *params, Ref fn, Operand engine);
 			virtual Nat STORM_FN firstParamId(MAYBE(TypeDesc *) desc);
 			virtual Operand STORM_FN firstParamLoc(Nat id);
+			virtual Reg STORM_FN functionDispatchReg();
 		};
 
 	}
