@@ -7,7 +7,23 @@ namespace code {
 	namespace x64 {
 
 		void prologOut(Output *to, Instr *instr) {
-			to->markProlog();
+			// We generate:
+			// - push rbp
+			// - mov rbp, rsp
+
+			// push rbp
+			to->putByte(0x50 + 5);
+			// now the CFA offset is different
+			to->setFrameOffset(Offset::sPtr*2);
+			// We also saved RBP on the stack
+			to->markSaved(ptrFrame, -Offset::sPtr*2);
+
+			// mov rbp, rsp
+			to->putByte(0x48);
+			to->putByte(0x89);
+			to->putByte(0xE5);
+			// now we use ebp as the CFA register, offset is the same
+			to->setFrameRegister(ptrFrame);
 		}
 
 		void epilogOut(Output *to, Instr *instr) {
@@ -16,16 +32,26 @@ namespace code {
 			// In general, the epilog CFA information is not strictly neccessary as long as there
 			// are no exceptions thrown during the last two instructions in the epilog.
 			// to->markEpilog();
+
+			// We use this pseudo-op to generate the LEAVE instruction.
+			to->putByte(0xC9);
 		}
 
 		void preserveOut(Output *to, Instr *instr) {
-			to->markSaved(instr->src().reg(), instr->dest().offset());
+			// Offset between the stack pointer and the CFA. This difference is due to us spilling
+			// RBP from the previous function, and the return address on the stack.
+			Size offset = Size::sPtr * 2;
+			to->markSaved(instr->src().reg(), instr->dest().offset() - offset);
 		}
 
 		void locationOut(Output *, Instr *) {
 			// We don't currently offer the ability to lookup source locations in the final binary.
 		}
 
+		void metaOut(Output *, Instr *) {
+			// We don't preserve metadata.
+		}
+
 		void pushOut(Output *to, Instr *instr) {
 			const Operand &src = instr->src();
 			switch (src.type()) {
@@ -278,6 +304,21 @@ namespace code {
 			immRegInstr(to, op8, op, instr->dest(), instr->src());
 		}
 
+		void testOut(Output *to, Instr *instr) {
+			ImmRegInstr8 op8 = {
+				opCode(0xF6), 0,
+				opCode(0x84),
+				opCode(0x84), // TEST is symmetric, order does not matter
+			};
+			ImmRegInstr op = {
+				opCode(0x00), 0xFF, // not supported
+				opCode(0xF7), 0,
+				opCode(0x85),
+				opCode(0x85), // TEST is symmetric, order does not matter
+			};
+			immRegInstr(to, op8, op, instr->dest(), instr->src());
+		}
+
 		static void shiftOp(Output *to, const Operand &dest, const Operand &src, byte subOp) {
 			RmFlags flags = wide(dest);
 			if (dest.size() == Size::sByte)
@@ -572,16 +613,6 @@ namespace code {
 			}
 		}
 
-		void fistpOut(Output *to, Instr *instr) {
-			// We're actually using the FISTTP instruction, since that fixes proper truncation for us!
-			// It does, however, require SSE3. But that seems fair.
-			if (instr->size() == Size::sLong) {
-				modRm(to, opCode(0xDD), rmNone, 1, instr->dest());
-			} else {
-				modRm(to, opCode(0xDB), rmNone, 1, instr->dest());
-			}
-		}
-
 		void fldOut(Output *to, Instr *instr) {
 			if (instr->size() == Size::sDouble) {
 				modRm(to, opCode(0xDD), rmNone, 0, instr->src());
@@ -590,52 +621,89 @@ namespace code {
 			}
 		}
 
-		void fildOut(Output *to, Instr *instr) {
-			if (instr->size() == Size::sLong) {
-				modRm(to, opCode(0xDF), rmNone, 5, instr->src());
-			} else {
-				modRm(to, opCode(0xDB), rmNone, 0, instr->src());
+		static void fpOp(Output *to, Instr *instr, OpCode op) {
+			if (instr->size() == Size::sDouble)
+				op.prefix = 0xF2;
+			else
+				op.prefix = 0xF3;
+			modRm(to, op, rmNone, instr->dest(), instr->src());
+		}
+
+		void faddOut(Output *to, Instr *instr) {
+			fpOp(to, instr, opCode(0x0F, 0x58));
+		}
+
+		void fsubOut(Output *to, Instr *instr) {
+			fpOp(to, instr, opCode(0x0F, 0x5C));
+		}
+
+		void fmulOut(Output *to, Instr *instr) {
+			fpOp(to, instr, opCode(0x0F, 0x59));
+		}
+
+		void fdivOut(Output *to, Instr *instr) {
+			fpOp(to, instr, opCode(0x0F, 0x5E));
+		}
+
+		void fnegOut(Output *to, Instr *instr) {
+			// set dest to zero, it has to be a register (using XORPS/XORPD)
+			Operand dest = instr->dest();
+			if (dest.size() == Size::sDouble)
+				modRm(to, prefixOpCode(0x66, 0x0F, 0x57), rmNone, dest, dest);
+			else
+				modRm(to, opCode(0x0F, 0x57), rmNone, dest, dest);
+			// fsub
+			fpOp(to, instr, opCode(0x0F, 0x5C));
+		}
+
+		void fcmpOut(Output *to, Instr *instr) {
+			// Note: this is COMISS/COMISD, not CMPSS/CMPSD
+			OpCode op = instr->size() == Size::sDouble
+				? prefixOpCode(0x66, 0x0F, 0x2F)
+				: opCode(0x0F, 0x2F);
+			modRm(to, op, rmNone, instr->dest(), instr->src());
+		}
+
+		void fcastOut(Output *to, Instr *instr) {
+			Operand src = instr->src();
+			Operand dst = instr->dest();
+
+			if (src.size() == dst.size()) {
+				// Just a mov operation!
+				movOut(to, instr);
+				return;
 			}
+
+			if (src.size() == Size::sDouble)
+				modRm(to, prefixOpCode(0xF2, 0x0F, 0x5A), rmNone, instr->dest(), instr->src());
+			else
+				modRm(to, prefixOpCode(0xF3, 0x0F, 0x5A), rmNone, instr->dest(), instr->src());
 		}
 
-		void fldzOut(Output *to, Instr *instr) {
-			to->putByte(0xD9);
-			to->putByte(0xEE);
+		void fcastiOut(Output *to, Instr *instr) {
+			byte prefix = instr->src().size() == Size::sDouble
+				? 0xF2 : 0xF3;
+			modRm(to, prefixOpCode(prefix, 0x0F, 0x2C), wide(instr->dest()), instr->dest(), instr->src());
 		}
 
-		void faddpOut(Output *to, Instr *instr) {
-			to->putByte(0xDE);
-			to->putByte(0xC1);
+		void fcastuOut(Output *to, Instr *instr) {
+			// TODO: Do something sensible!
+			fcastiOut(to, instr);
 		}
 
-		void fsubpOut(Output *to, Instr *instr) {
-			to->putByte(0xDE);
-			to->putByte(0xE9);
+		void icastfOut(Output *to, Instr *instr) {
+			// TODO: Need to clear target register first? (use PXOR, 66 0f ef XX)
+
+			byte prefix = instr->dest().size() == Size::sDouble
+				? 0xF2 : 0xF3;
+			modRm(to, prefixOpCode(prefix, 0x0F, 0x2A), wide(instr->src()), instr->dest(), instr->src());
 		}
 
-		void fmulpOut(Output *to, Instr *instr) {
-			to->putByte(0xDE);
-			to->putByte(0xC9);
+		void ucastfOut(Output *to, Instr *instr) {
+			// TODO: Do something sensible!
+			icastfOut(to, instr);
 		}
 
-		void fdivpOut(Output *to, Instr *instr) {
-			to->putByte(0xDE);
-			to->putByte(0xF9);
-		}
-
-		void fcomppOut(Output *to, Instr *instr) {
-			// fcomip ST1 (sets EFLAGS)
-			to->putByte(0xDF);
-			to->putByte(0xF0 + 1);
-
-			// fstp ST0 (effectively only a pop)
-			to->putByte(0xDD);
-			to->putByte(0xD8 + 0);
-		}
-
-		void fwaitOut(Output *to, Instr *instr) {
-			to->putByte(0x9B);
-		}
 
 		void datOut(Output *to, Instr *instr) {
 			Operand src = instr->src();
@@ -684,6 +752,7 @@ namespace code {
 			OUTPUT(sbb),
 			OUTPUT(bxor),
 			OUTPUT(cmp),
+			OUTPUT(test),
 			OUTPUT(shl),
 			OUTPUT(shr),
 			OUTPUT(sar),
@@ -705,19 +774,22 @@ namespace code {
 			OUTPUT(epilog),
 			OUTPUT(preserve),
 			OUTPUT(location),
+			OUTPUT(meta),
 
 			// Floating point.
 			OUTPUT(fstp),
-			OUTPUT(fistp),
 			OUTPUT(fld),
-			OUTPUT(fild),
-			OUTPUT(fldz),
-			OUTPUT(faddp),
-			OUTPUT(fsubp),
-			OUTPUT(fmulp),
-			OUTPUT(fdivp),
-			OUTPUT(fcompp),
-			OUTPUT(fwait),
+			OUTPUT(fadd),
+			OUTPUT(fsub),
+			OUTPUT(fmul),
+			OUTPUT(fdiv),
+			OUTPUT(fneg),
+			OUTPUT(fcmp),
+			OUTPUT(fcast),
+			OUTPUT(fcasti),
+			OUTPUT(fcastu),
+			OUTPUT(icastf),
+			OUTPUT(ucastf),
 
 			OUTPUT(dat),
 			OUTPUT(lblOffset),
