@@ -4,6 +4,7 @@
 #include "Asm.h"
 #include "../Binary.h"
 #include "../Layout.h"
+#include "../PosixEh/FnState.h"
 
 namespace code {
 	namespace x64 {
@@ -25,46 +26,30 @@ namespace code {
 			TRANSFORM(fnRetRef),
 		};
 
-		Layout::Active::Active(Block block, Nat activated, Label pos) : block(block), activated(activated), pos(pos) {}
 
-		Layout::Layout(Binary *owner) : owner(owner) {}
+		Layout::Layout() {}
 
 		void Layout::before(Listing *dest, Listing *src) {
 			// Initialize some state.
 			block = Block();
 			usingEH = src->exceptionAware();
 
-			// Compute where the result is to be stored.
-			result = code::x64::result(src->result);
-
 			// Compute the layout of all parameters.
 			params = new (this) Params();
-
-			// Hidden return parameter?
-			if (result->memory)
-				params->add(Param::returnId, ptrPrimitive());
-
-			// Add the rest of them.
+			params->result(src->result);
 			Array<Var> *p = src->allParams();
 			for (Nat i = 0; i < p->count(); i++) {
 				params->add(i, src->paramDesc(p->at(i)));
 			}
 
 			// Figure out which registers we need to spill.
-			{
-				toPreserve = new (this) RegSet();
-				RegSet *notPreserved = fnDirtyRegs(engine());
-				RegSet *used = allUsedRegs(src);
-
-				for (RegSet::Iter i = used->begin(); i != used->end(); ++i) {
-					if (!notPreserved->has(i.v()))
-						toPreserve->put(i.v());
-				}
-			}
+			toPreserve = allUsedRegs(src);
+			for (size_t i = 0; i < fnDirtyCount; i++)
+				toPreserve->remove(fnDirtyRegs[i]);
 
 			// Compute the layout.
 			Nat spilled = toPreserve->count();
-			if (result->memory)
+			if (params->result().memoryRegister() != noReg)
 				spilled++; // The hidden parameter needs to be spilled!
 
 			layout = code::x64::layout(src, params, spilled);
@@ -81,7 +66,7 @@ namespace code {
 			}
 
 			// The EH table.
-			activeBlocks = new (this) Array<Active>();
+			activeBlocks = new (this) Array<ActiveBlock>();
 		}
 
 		void Layout::during(Listing *dest, Listing *src, Nat line) {
@@ -100,8 +85,8 @@ namespace code {
 			*dest << alignAs(Size::sPtr);
 			*dest << dest->meta();
 
-			// Total stack size.
-			*dest << dat(ptrConst(layout->last()));
+			// Offset between RBP and RSP.
+			*dest << dat(ptrConst(-layout->last()));
 
 			// Output metadata table.
 			Array<Var> *vars = src->allVars();
@@ -129,15 +114,13 @@ namespace code {
 			*dest << alignAs(Size::sPtr);
 			// Table contents. Each 'row' is 8 bytes.
 			for (Nat i = 0; i < activeBlocks->count(); i++) {
-				const Active &a = activeBlocks->at(i);
+				const ActiveBlock &a = activeBlocks->at(i);
 				*dest << lblOffset(a.pos);
-				*dest << dat(natConst(a.encode()));
+				*dest << dat(natConst(code::eh::encodeFnState(a.block.key(), a.activated)));
 			}
 
 			// Table size.
 			*dest << dat(ptrConst(activeBlocks->count()));
-			// Owner.
-			*dest << dat(objPtr(owner));
 		}
 
 		Operand Layout::resolve(Listing *src, const Operand &op) {
@@ -158,10 +141,10 @@ namespace code {
 			Array<Var> *all = dest->allParams();
 
 			for (Nat i = 0; i < params->registerCount(); i++) {
-				Param info = params->registerAt(i);
-				if (info.size() == 0)
+				Param info = params->registerParam(i);
+				if (info.size() == Size())
 					continue; // Not used.
-				if (info.id() == Param::returnId) {
+				if (info.id() == Param::returnId()) {
 					*dest << mov(ptrRel(ptrFrame, resultParam()), params->registerSrc(i));
 					continue;
 				}
@@ -181,12 +164,13 @@ namespace code {
 		}
 
 		void Layout::prologTfm(Listing *dest, Listing *src, Nat line) {
-			*dest << push(ptrFrame);
-			*dest << mov(ptrFrame, ptrStack);
-
-			// Notify that we've generated the prolog.
+			// Generate the prolog. Generates push and mov to set up a basic stack frame. Also emits
+			// proper unwind data.
 			*dest << prolog();
 
+			// *dest << push(ptrFrame);
+			// *dest << mov(ptrFrame, ptrStack);
+
 			// Allocate stack space.
 			if (layout->last() != Offset())
 				*dest << sub(ptrStack, ptrConst(layout->last()));
@@ -205,15 +189,19 @@ namespace code {
 			spillParams(dest);
 
 			// Initialize the root block.
-			initBlock(dest, dest->root());
+			initBlock(dest, dest->root(), rax);
 		}
 
 		void Layout::epilogTfm(Listing *dest, Listing *src, Nat line) {
+			epilog(dest, src, line, true);
+		}
+
+		void Layout::epilog(Listing *dest, Listing *src, Nat line, Bool preserveRax) {
 			// Destroy blocks. Note: we shall not modify 'block' nor alter the exception table as
 			// this may be an early return from the function.
 			Block oldBlock = block;
 			for (Block now = block; now != Block(); now = src->parent(now)) {
-				destroyBlock(dest, now, true, false);
+				destroyBlock(dest, now, preserveRax, false);
 			}
 			block = oldBlock;
 
@@ -224,15 +212,21 @@ namespace code {
 				offset -= Offset::sPtr;
 			}
 
-			*dest << mov(ptrStack, ptrFrame);
-			*dest << pop(ptrFrame);
+			// The "epilog" pseudo-op generates a LEAVE instruction, which corresponds to these instructions:
+			// *dest << mov(ptrStack, ptrFrame);
+			// *dest << pop(ptrFrame);
 
-			// Notify that we've generated the epilog.
-			*dest << epilog();
+			*dest << code::epilog();
 		}
 
 		void Layout::beginBlockTfm(Listing *dest, Listing *src, Nat line) {
-			initBlock(dest, src->at(line)->src().block());
+			Instr *instr = src->at(line);
+			// Note: register is added in the previous pass.
+			Reg freeReg = noReg;
+			if (instr->dest().type() == opRegister)
+				freeReg = instr->dest().reg();
+
+			initBlock(dest, instr->src().block(), freeReg);
 		}
 
 		void Layout::endBlockTfm(Listing *dest, Listing *src, Nat line) {
@@ -273,7 +267,7 @@ namespace code {
 			if (src->freeOpt(var) & freeOnException) {
 				Label lbl = dest->label();
 				*dest << lbl;
-				activeBlocks->push(Active(block, activationId, lbl));
+				activeBlocks->push(ActiveBlock(block, activationId, lbl));
 			}
 		}
 
@@ -282,17 +276,19 @@ namespace code {
 			return -(Offset::sPtr * count);
 		}
 
-		// Zero the memory of a variable. 'initEax' should be true if we need to set eax to 0 before
-		// using it as our zero. 'initRax' will be set to false, so that it is easy to use zeroVar
-		// in a loop, causing only the first invocation to emit 'rax := 0'.
-		static void zeroVar(Listing *dest, Offset start, Size size, Bool &initRax) {
+		// Zero the memory of a variable. 'initReg' should be true if we need to set <reg> to 0
+		// before using it as our zero. 'initReg' will be set to false, so that it is easy to use
+		// zeroVar in a loop, causing only the first invocation to emit '<reg> := 0'.
+		static void zeroVar(Listing *dest, Offset start, Size size, Reg reg, Bool &initReg) {
 			nat s64 = size.size64();
 			if (s64 == 0)
 				return;
 
-			if (initRax) {
-				*dest << bxor(rax, rax);
-				initRax = false;
+			reg = asSize(reg, Size::sLong);
+
+			if (initReg) {
+				*dest << bxor(reg, reg);
+				initReg = false;
 			}
 
 			// Note: We need to be careful not to overshoot too much. We might have a stack of
@@ -301,19 +297,19 @@ namespace code {
 			nat pos = 0;
 			while (pos < s64) {
 				if (s64 - pos >= 8) {
-					*dest << mov(longRel(ptrFrame, start + Offset(pos)), rax);
+					*dest << mov(longRel(ptrFrame, start + Offset(pos)), reg);
 					pos += 8;
 				} else if (s64 - pos >= 4) {
-					*dest << mov(intRel(ptrFrame, start + Offset(pos)), eax);
+					*dest << mov(intRel(ptrFrame, start + Offset(pos)), asSize(reg, Size::sInt));
 					pos += 4;
 				} else {
-					*dest << mov(byteRel(ptrFrame, start + Offset(pos)), al);
+					*dest << mov(byteRel(ptrFrame, start + Offset(pos)), asSize(reg, Size::sByte));
 					pos += 1;
 				}
 			}
 		}
 
-		void Layout::initBlock(Listing *dest, Block init) {
+		void Layout::initBlock(Listing *dest, Block init, Reg reg) {
 			if (block != dest->parent(init)) {
 				Str *msg = TO_S(engine(), S("Can not begin ") << init << S(" unless the current is ")
 								<< dest->parent(init) << S(". Current is ") << block);
@@ -322,21 +318,57 @@ namespace code {
 
 			block = init;
 
-			Bool initEax = true;
+			Bool initReg = true;
+			Bool restoreReg = reg == noReg;
+			if (restoreReg) {
+				reg = eax;
+				*dest << push(eax);
+			}
+
 			Array<Var> *vars = dest->allVars(init);
 			// Go in reverse to make linear accesses in memory when we're using big variables.
 			for (Nat i = vars->count(); i > 0; i--) {
 				Var v = vars->at(i - 1);
 
 				if (!dest->isParam(v))
-					zeroVar(dest, layout->at(v.key()), v.size(), initEax);
+					zeroVar(dest, layout->at(v.key()), v.size(), reg, initReg);
 			}
 
+			if (restoreReg)
+				*dest << pop(eax);
+
 			if (usingEH) {
 				// Remember where the block started.
 				Label lbl = dest->label();
 				*dest << lbl;
-				activeBlocks->push(Active(block, activationId, lbl));
+				activeBlocks->push(ActiveBlock(block, activationId, lbl));
+			}
+		}
+
+		static void saveResult(Listing *dest, const Result &result) {
+			if (result.memoryRegister() != noReg) {
+				// We need to keep the stack aligned to 16 bits.
+				*dest << push(result.memoryRegister());
+				*dest << push(result.memoryRegister());
+			} else if (result.registerCount() > 0) {
+				Size sz = Size::sPtr * roundUp(result.registerCount(), Nat(2));
+				*dest << sub(ptrStack, ptrConst(sz));
+				for (Nat i = 0; i < result.registerCount(); i++) {
+					*dest << mov(ptrRel(ptrStack, Offset::sPtr * i), asSize(result.registerAt(i), Size::sPtr));
+				}
+			}
+		}
+
+		static void restoreResult(Listing *dest, const Result &result) {
+			if (result.memoryRegister() != noReg) {
+				*dest << pop(result.memoryRegister());
+				*dest << pop(result.memoryRegister());
+			} else if (result.registerCount() > 0) {
+				Size sz = Size::sPtr * roundUp(result.registerCount(), Nat(2));
+				for (Nat i = 0; i < result.registerCount(); i++) {
+					*dest << mov(asSize(result.registerAt(i), Size::sPtr), ptrRel(ptrStack, Offset::sPtr * i));
+				}
+				*dest << add(ptrStack, ptrConst(sz));
 			}
 		}
 
@@ -344,7 +376,7 @@ namespace code {
 			if (destroy != block)
 				throw new (this) BlockEndError();
 
-			Bool pushedRax = false;
+			Bool pushedResult = false;
 			Array<Var> *vars = dest->allVars(destroy);
 			// Destroy in reverse order.
 			for (Nat i = vars->count(); i > 0; i--) {
@@ -358,9 +390,9 @@ namespace code {
 					if (activated->at(v.key()) > activationId)
 						continue;
 
-					if (preserveRax && !pushedRax) {
-						saveResult(dest, dest->result);
-						pushedRax = true;
+					if (preserveRax && !pushedResult) {
+						saveResult(dest, params->result());
+						pushedResult = true;
 					}
 
 					if (when & freeIndirection) {
@@ -385,32 +417,15 @@ namespace code {
 				}
 			}
 
-			if (pushedRax) {
-				restoreResult(dest, dest->result);
+			if (pushedResult) {
+				restoreResult(dest, params->result());
 			}
 
 			block = dest->parent(block);
 			if (usingEH && table) {
 				Label lbl = dest->label();
 				*dest << lbl;
-				activeBlocks->push(Active(block, activationId, lbl));
-			}
-		}
-
-		static void returnLayout(Listing *dest, primitive::PrimitiveKind k, nat &i, nat &r, Offset offset) {
-			static const Reg intReg[2] = { ptrA, ptrD };
-			static const Reg realReg[2] = { xmm0, xmm1 };
-
-			switch (k) {
-			case primitive::none:
-				break;
-			case primitive::integer:
-			case primitive::pointer:
-				*dest << mov(intReg[i++], ptrRel(ptrSi, offset));
-				break;
-			case primitive::real:
-				*dest << mov(realReg[r++], longRel(ptrSi, offset));
-				break;
+				activeBlocks->push(ActiveBlock(block, activationId, lbl));
 			}
 		}
 
@@ -436,16 +451,16 @@ namespace code {
 		}
 
 		// Put the return value into registers. Assumes ptrSi contains a pointer to the struct to be returned.
-		static void returnSimple(Listing *dest, Result *result, SimpleDesc *type, Offset resultParam) {
-			if (result->memory) {
-				*dest << mov(ptrA, ptrRel(ptrFrame, resultParam));
+		static void returnSimple(Listing *dest, const Result &result, SimpleDesc *type, Offset resultParam) {
+			if (result.memoryRegister() != noReg) {
+				*dest << mov(result.memoryRegister(), ptrRel(ptrFrame, resultParam));
 				// Inline a memcpy using mov instructions.
-				movMemcpy(dest, ptrA, ptrSi, type->size());
+				movMemcpy(dest, result.memoryRegister(), ptrSi, type->size());
 			} else {
-				nat i = 0;
-				nat r = 0;
-				returnLayout(dest, result->part1, i, r, Offset());
-				returnLayout(dest, result->part2, i, r, Offset::sPtr);
+				for (Nat i = 0; i < result.registerCount(); i++) {
+					Reg r = result.registerAt(i);
+					*dest << mov(r, xRel(size(r), ptrSi, Offset(result.registerOffset(i))));
+				}
 			}
 		}
 
@@ -471,7 +486,14 @@ namespace code {
 
 		void Layout::fnRetTfm(Listing *dest, Listing *src, Nat line) {
 			Operand value = resolve(src, src->at(line)->src());
-			assert(value.size() == src->result->size(), L"Wrong size passed to fnRet!");
+			if (value.size() != dest->result->size()) {
+				StrBuf *msg = new (this) StrBuf();
+				*msg << S("Wrong size passed to fnRet. Got: ");
+				*msg << value.size();
+				*msg << S(" but expected ");
+				*msg << dest->result->size() << S(".");
+				throw new (this) InvalidValue(msg->toS());
+			}
 
 			// Handle the return value.
 			if (PrimitiveDesc *p = as<PrimitiveDesc>(src->result)) {
@@ -485,7 +507,7 @@ namespace code {
 				*dest << mov(ptrA, ptrRel(ptrFrame, resultParam()));
 			} else if (SimpleDesc *s = as<SimpleDesc>(src->result)) {
 				*dest << lea(ptrSi, value);
-				returnSimple(dest, result, s, resultParam());
+				returnSimple(dest, params->result(), s, resultParam());
 			} else {
 				assert(false);
 			}
@@ -527,7 +549,7 @@ namespace code {
 				*dest << mov(ptrA, ptrRel(ptrFrame, resultParam()));
 			} else if (SimpleDesc *s = as<SimpleDesc>(src->result)) {
 				*dest << mov(ptrSi, value);
-				returnSimple(dest, result, s, resultParam());
+				returnSimple(dest, params->result(), s, resultParam());
 			} else {
 				assert(false);
 			}
@@ -549,7 +571,7 @@ namespace code {
 				// Start by computing the locations of parameters passed on the stack.
 				Offset stackOffset = Offset::sPtr * 2;
 				for (Nat i = 0; i < params->stackCount(); i++) {
-					Nat paramId = params->stackAt(i); // Should never be 'Param::returnId'
+					Nat paramId = params->stackParam(i).id(); // Should never be 'Param::returnId'
 					Var var = all->at(paramId);
 
 					out->at(var.key()) = stackOffset;
