@@ -1,14 +1,14 @@
 #pragma once
 #include "../Output.h"
 #include "Core/GcCode.h"
-#include "DwarfEh.h"
+#include "Code/Dwarf/CodeOutput.h"
 
 namespace code {
 	class Binary;
 	namespace x64 {
 		STORM_PKG(core.asm.x64);
 
-		class CodeOut : public CodeOutput {
+		class CodeOut : public code::dwarf::CodeOutput {
 			STORM_CLASS;
 		public:
 			STORM_CTOR CodeOut(Binary *owner, Array<Nat> *lbls, Nat size, Nat refs);
@@ -19,16 +19,13 @@ namespace code {
 			virtual void STORM_FN putPtr(Word w);
 			virtual void STORM_FN align(Nat to);
 			virtual void putGc(GcCodeRef::Kind kind, Nat size, Word w);
+			virtual void markGc(GcCodeRef::Kind kind, Nat size, Word w);
 			virtual void STORM_FN putGcPtr(Word w);
 			virtual void STORM_FN putGcRelative(Word w);
 			virtual void STORM_FN putRelativeStatic(Word w);
 			virtual void STORM_FN putPtrSelf(Word w);
 			virtual Nat STORM_FN tell() const;
 
-			virtual void STORM_FN markProlog();
-			virtual void STORM_FN markEpilog();
-			virtual void STORM_FN markSaved(Reg reg, Offset offset);
-
 			virtual void *codePtr() const;
 		protected:
 			virtual void STORM_FN markLabel(Nat id);
@@ -46,12 +43,6 @@ namespace code {
 			// Code we're writing to.
 			UNKNOWN(PTR_GC) byte *code;
 
-			// CFI object we're writing to.
-			FnInfo fnInfo;
-
-			// Position in the code.
-			Nat pos;
-
 			// Size of the code.
 			Nat size;
 
