@@ -6,11 +6,79 @@
 namespace code {
 	namespace x86 {
 
-#define OUTPUT(x) { op::x, &x ## Out }
+		void epilogOut(Output *to, Instr *instr) {
+			(void)instr;
+			// This is LEAVE.
+			to->putByte(0xC9);
+		}
 
-		typedef void (*OutputFn)(Output *to, Instr *instr);
+		static void movFromFp(Output *to, Instr *instr) {
+			Operand dst = instr->dest();
+			bool large = dst.size().size32() > 4;
+			bool spilled = false;
+			if (dst.type() == opRegister) {
+				// Need to spill to memory.
+				spilled = true;
+				dst = xRel(dst.size(), ptrStack, Offset());
+				// Reserve memory on stack (esp -= 4/8)
+				to->putByte(0x81);
+				modRm(to, 5, ptrStack);
+				to->putInt(dst.size().size32());
+			}
 
-		void movOut(Output *to, Instr *instr) {
+			if (large)
+				to->putByte(0xF2);
+			else
+				to->putByte(0xF3);
+			to->putByte(0x0F);
+			to->putByte(0x11);
+			modRm(to, fpRegisterId(instr->src().reg()), dst);
+
+			if (spilled) {
+				dst = instr->dest();
+				if (large) {
+					to->putByte(0x58 + registerId(low32(dst.reg())));
+					to->putByte(0x58 + registerId(high32(dst.reg())));
+				} else {
+					to->putByte(0x58 + registerId(dst.reg()));
+				}
+			}
+		}
+
+		static void movToFp(Output *to, Instr *instr) {
+			Operand src = instr->src();
+			bool large = src.size().size32() > 4;
+			bool spilled = false;
+			if (src.type() == opRegister) {
+				// Spill to memory using push.
+				spilled = true;
+				if (large) {
+					to->putByte(0x50 + registerId(high32(src.reg())));
+					to->putByte(0x50 + registerId(low32(src.reg())));
+				} else {
+					to->putByte(0x50 + registerId(src.reg()));
+				}
+				src = xRel(src.size(), ptrStack, Offset());
+			}
+
+			if (large)
+				to->putByte(0xF2);
+			else
+				to->putByte(0xF3);
+			to->putByte(0x0F);
+			to->putByte(0x10);
+			modRm(to, fpRegisterId(instr->dest().reg()), src);
+
+			if (spilled) {
+				// Clear stack (esp += 4/8)
+				to->putByte(0x81);
+				modRm(to, 0, ptrStack);
+				to->putInt(src.size().size32());
+			}
+		}
+
+		static void movOut(Output *to, const Operand &dest, const Operand &src) {
+			// Integer mov.
 			ImmRegInstr8 op8 = {
 				0xC6, 0,
 				0x88,
@@ -22,7 +90,28 @@ namespace code {
 				0x89,
 				0x8B,
 			};
-			immRegInstr(to, op8, op, instr->dest(), instr->src());
+			immRegInstr(to, op8, op, dest, src);
+		}
+
+		void movOut(Output *to, Instr *instr) {
+			bool fpSrc = fpRegister(instr->src());
+			bool fpDest = fpRegister(instr->dest());
+			if (fpSrc && fpDest) {
+				// Two fp registers:
+				if (instr->size() == Size::sDouble)
+					to->putByte(0xF2);
+				else
+					to->putByte(0xF3);
+				to->putByte(0x0F);
+				to->putByte(0x10);
+				modRm(to, fpRegisterId(instr->dest().reg()), instr->src());
+			} else if (fpSrc) {
+				movFromFp(to, instr);
+			} else if (fpDest) {
+				movToFp(to, instr);
+			} else {
+				movOut(to, instr->dest(), instr->src());
+			}
 		}
 
 		void swapOut(Output *to, Instr *instr) {
@@ -34,7 +123,7 @@ namespace code {
 			modRm(to, registerId(instr->dest().reg()), instr->src());
 		}
 
-		void addOut(Output *to, Instr *instr) {
+		static void addOut(Output *to, const Operand &dest, const Operand &src) {
 			ImmRegInstr8 op8 = {
 				0x80, 0,
 				0x00,
@@ -46,10 +135,13 @@ namespace code {
 				0x01,
 				0x03
 			};
-			immRegInstr(to, op8, op, instr->dest(), instr->src());
+			immRegInstr(to, op8, op, dest, src);
+		}
+		void addOut(Output *to, Instr *instr) {
+			addOut(to, instr->dest(), instr->src());
 		}
 
-		void adcOut(Output *to, Instr *instr) {
+		static void adcOut(Output *to, const Operand &dest, const Operand &src) {
 			ImmRegInstr8 op8 = {
 				0x80, 2,
 				0x10,
@@ -61,7 +153,10 @@ namespace code {
 				0x11,
 				0x13
 			};
-			immRegInstr(to, op8, op, instr->dest(), instr->src());
+			immRegInstr(to, op8, op, dest, src);
+		}
+		void adcOut(Output *to, Instr *instr) {
+			adcOut(to, instr->dest(), instr->src());
 		}
 
 		void borOut(Output *to, Instr *instr) {
@@ -137,7 +232,7 @@ namespace code {
 			immRegInstr(to, op8, op, instr->dest(), instr->src());
 		}
 
-		void bxorOut(Output *to, Instr *instr) {
+		static void bxorOut(Output *to, const Operand &dest, const Operand &src) {
 			ImmRegInstr8 op8 = {
 				0x80, 6,
 				0x30,
@@ -149,7 +244,10 @@ namespace code {
 				0x31,
 				0x33
 			};
-			immRegInstr(to, op8, op, instr->dest(), instr->src());
+			immRegInstr(to, op8, op, dest, src);
+		}
+		void bxorOut(Output *to, Instr *instr) {
+			bxorOut(to, instr->dest(), instr->src());
 		}
 
 		void cmpOut(Output *to, Instr *instr) {
@@ -167,6 +265,21 @@ namespace code {
 			immRegInstr(to, op8, op, instr->dest(), instr->src());
 		}
 
+		void testOut(Output *to, Instr *instr) {
+			ImmRegInstr8 op8 = {
+				0xF6, 7,
+				0x84,
+				0x84 // TEST is symmetric
+			};
+			ImmRegInstr op = {
+				0x0, 0xFF, // Not supported
+				0xF7, 0,
+				0x85,
+				0x85 // TEST is symmetric
+			};
+			immRegInstr(to, op8, op, instr->dest(), instr->src());
+		}
+
 		void pushOut(Output *to, Instr *instr) {
 			const Operand &src = instr->src();
 			assert(src.size() != Size::sLong);
@@ -493,115 +606,327 @@ namespace code {
 		}
 
 		void fstpOut(Output *to, Instr *instr) {
-			if (instr->size() == Size::sDouble)
+			if (instr->dest().size() == Size::sDouble)
 				to->putByte(0xDD);
 			else
 				to->putByte(0xD9);
 			modRm(to, 3, instr->dest());
 		}
 
-		void fistpOut(Output *to, Instr *instr) {
-#ifdef X86_REQUIRE_SSE3
-			// Issue FISTTP instead.
-			if (instr->size() == Size::sLong) {
-				to->putByte(0xDD);
-				modRm(to, 1, instr->dest());
-			} else {
-				to->putByte(0xDB);
-				modRm(to, 1, instr->dest());
-			}
-#else
-			// Use space just above stack for this.
-			Operand modified = intRel(ptrStack, -Offset(4));
-			Operand old = intRel(ptrStack, -Offset(2));
-
-			// Set rounding mode to 'truncate'.
-
-			// FNSTCW [ESP - 2]
-			to->putByte(0xD9);
-			modRm(to, 7, old);
-
-			// FNSTCW [ESP - 4]
-			to->putByte(0xD9);
-			modRm(to, 7, modified);
-
-			// OR [esp - 4], #0xC00 - Set bits 10 and 11 to 1.
-			borOut(to, bor(instr->engine(), modified, natConst(0xC00)));
-
-			// FLDCW [ESP - 4]
-			to->putByte(0xD9);
-			modRm(to, 5, modified);
-
-			// FISTP 'dest'
-			if (instr->size() == Size::sLong) {
-				to->putByte(0xDF);
-				modRm(to, 7, instr->dest());
-			} else {
-				to->putByte(0xDB);
-				modRm(to, 3, instr->dest());
-			}
-
-			// FLDCW [ESP - 2] - restore old mode
-			to->putByte(0xD9);
-			modRm(to, 5, old);
-#endif
-		}
-
 		void fldOut(Output *to, Instr *instr) {
-			if (instr->size() == Size::sDouble)
+			if (instr->src().size() == Size::sDouble)
 				to->putByte(0xDD);
 			else
 				to->putByte(0xD9);
 			modRm(to, 0, instr->src());
 		}
 
-		void fildOut(Output *to, Instr *instr) {
-			if (instr->size() == Size::sLong) {
-				to->putByte(0xDF);
-				modRm(to, 5, instr->src());
+		static void fpOp(Output *to, Instr *instr, byte op) {
+			if (instr->size() == Size::sDouble)
+				to->putByte(0xF2);
+			else
+				to->putByte(0xF3);
+			to->putByte(0x0F);
+			to->putByte(op);
+			modRm(to, fpRegisterId(instr->dest().reg()), instr->src());
+		}
+
+		void faddOut(Output *to, Instr *instr) {
+			fpOp(to, instr, 0x58);
+		}
+
+		void fsubOut(Output *to, Instr *instr) {
+			fpOp(to, instr, 0x5C);
+		}
+
+		void fmulOut(Output *to, Instr *instr) {
+			fpOp(to, instr, 0x59);
+		}
+
+		void fdivOut(Output *to, Instr *instr) {
+			fpOp(to, instr, 0x5E);
+		}
+
+		void fnegOut(Output *to, Instr *instr) {
+			// set dest to zero, it has to be a register (using XORPS/XORPD)
+			Operand dest = instr->dest();
+			if (dest.size() == Size::sDouble)
+				to->putByte(0x66);
+			to->putByte(0x0F);
+			to->putByte(0x57);
+			modRm(to, fpRegisterId(dest.reg()), dest);
+
+			// fsub
+			fpOp(to, instr, 0x5C);
+		}
+
+		void fcmpOut(Output *to, Instr *instr) {
+			// Note: this is COMISS/COMISD, not CMPSS/CMPSD
+			if (instr->size() == Size::sDouble)
+				to->putByte(0x66);
+			to->putByte(0x0F);
+			to->putByte(0x2F);
+			modRm(to, fpRegisterId(instr->dest().reg()), instr->src());
+		}
+
+		void fcastOut(Output *to, Instr *instr) {
+			Operand src = instr->src();
+			Operand dst = instr->dest();
+
+			if (src.size() == dst.size()) {
+				// Just a mov operation!
+				movOut(to, instr);
+				return;
+			}
+
+			if (src.size() == Size::sDouble)
+				to->putByte(0xF2);
+			else
+				to->putByte(0xF3);
+			to->putByte(0x0F);
+			to->putByte(0x5A);
+			modRm(to, fpRegisterId(instr->dest().reg()), instr->src());
+		}
+
+		void fcastiOut(Output *to, Instr *instr) {
+			Operand src = instr->src();
+			Operand dst = instr->dest();
+			if (dst.size() == Size::sLong) {
+				// Use "old-style" FP instructions. We know that both operands are in memory, so we
+				// can simply emit fld and fisttp:
+				fldOut(to, instr);
+				to->putByte(0xDD);
+				modRm(to, 1, instr->dest());
 			} else {
-				to->putByte(0xDB);
-				modRm(to, 0, instr->src());
+				// Use CVTTSS2SI or CVTTSD2SI.
+				if (src.size() == Size::sDouble)
+					to->putByte(0xF2);
+				else
+					to->putByte(0xF3);
+				to->putByte(0x0F);
+				to->putByte(0x2C);
+				modRm(to, registerId(instr->dest().reg()), instr->src());
 			}
 		}
 
-		void fldzOut(Output *to, Instr *instr) {
-			to->putByte(0xD9);
-			to->putByte(0xEE);
+		void fcastuOut(Output *to, Instr *instr) {
+			// Operand src = instr->src();
+			Operand dst = instr->dest();
+
+			// Use "old-style" FP instructions. We know that both operands are in memory, so we
+			// can simply emit fld and fisttp:
+			fldOut(to, instr);
+
+			if (dst.size() == Size::sLong) {
+				// Complex case. Conversions to integers are 'staurated', and we use that. We spill
+				// the number to memory first, load it back and subtract from the original
+				// number. That gives us a difference that we can add to the original number.
+				// We know that we have 16 bytes available on the stack to do our work, and a register
+				// <dsth>, <dstl> to use as a temporary.
+
+				Reg dstl = low32(dst.reg());
+				Reg dsth = high32(dst.reg());
+				Operand stackPos = longRel(ptrStack, Offset());
+				Operand stackLow = low32(stackPos);
+				Operand stackHigh = high32(stackPos);
+
+
+				// We emit the following code: (source is already loaded)
+				// mov <dstl>, 0xFFFFFFFF
+				// mov <stack>, <dstl>
+				// shr <dstl>, 1
+				// mov <stack+4>, <dstl>
+				// fild <stack>
+				// fcomi ST1
+				// ja LARGE
+				// fldz ST1
+				// fcomip ST1
+				// ja NORMAL
+				// xor <dsth>, <dsth>
+				// xor <dstl>, <dstl>
+				// j DONE
+				// LARGE:
+				// fsubp ST1 (maybe fsubr)
+				// fsttp <stack>
+				// mov <dstl>, <stack>
+				// mov <dsth>, <stack+4>
+				// add <dstl>, 0xFFFFFFFF
+				// adc <dsth>, 0x7FFFFFFF
+				// j DONE
+				// NORMAL:
+				// fsttp <stack>
+				// mov <dstl>, <stack>
+				// mov <dsth>, <stack+4>
+				// DONE:
+
+				// mov <dstl>, 0xFFFFFF
+				movOut(to, dstl, natConst(0xFFFFFFFF));
+				// mov <stack>, <dstl>
+				movOut(to, stackLow, dstl);
+				// shr <dstl>
+				to->putByte(0xC1);
+				modRm(to, 5, dstl);
+				to->putByte(0x01);
+				// mov <stack+4>, <dstl>
+				movOut(to, stackHigh, dstl);
+				// fild
+				to->putByte(0xDF);
+				modRm(to, 5, stackPos);
+				// fcomi ST1
+				to->putByte(0xDB);
+				to->putByte(0xF1);
+				// ja LARGE
+				to->putByte(0x0F);
+				to->putByte(0x82);
+				to->putInt(23);
+				// fstp ST0 (effectively only a pop)
+				to->putByte(0xDD);
+				to->putByte(0xD8 + 0);
+				// fldz ST1
+				to->putByte(0xD9);
+				to->putByte(0xEE);
+				// fcomip ST1
+				to->putByte(0xDF);
+				to->putByte(0xF1);
+				// ja NORMAL
+				to->putByte(0x0F);
+				to->putByte(0x82);
+				to->putInt(40);
+				// fstp ST0 (effectively only a pop)
+				to->putByte(0xDD);
+				to->putByte(0xD8 + 0);
+				// xor x2 (4 bytes)
+				bxorOut(to, dstl, dstl);
+				bxorOut(to, dsth, dsth);
+				// j DONE
+				to->putByte(0xE9);
+				to->putInt(0x27);
+				// LARGE:
+				// fsubp/fsubr ST0
+				to->putByte(0xDE);
+				to->putByte(0xE9);
+				// fsttp <stack>
+				to->putByte(0xDD);
+				modRm(to, 1, stackPos);
+				// mov x2
+				movOut(to, dstl, stackLow);
+				movOut(to, dsth, stackHigh);
+				// add <dstl>, <stack>
+				addOut(to, dstl, natConst(0xFFFFFFFF));
+				// adc <dsth>, <stack+4>
+				adcOut(to, dsth, natConst(0x7FFFFFFF));
+				// j DONE
+				to->putByte(0xE9);
+				to->putInt(0xA);
+				// NORMAL:
+				// fsttp <stack>
+				to->putByte(0xDD);
+				modRm(to, 1, stackPos);
+				// mov x2
+				movOut(to, dstl, stackLow);
+				movOut(to, dsth, stackHigh);
+				// DONE:
+
+			} else {
+				// Simple case, just tell the hardware to emit a 64-bit output and return the lower
+				// 32-bits of it.
+				to->putByte(0xDD);
+				modRm(to, 1, instr->dest());
+			}
 		}
 
-		void faddpOut(Output *to, Instr *instr) {
-			to->putByte(0xDE);
-			to->putByte(0xC1);
+		void icastfOut(Output *to, Instr *instr) {
+			Operand src = instr->src();
+			// Operand dst = instr->dest();
+			if (src.size() == Size::sLong) {
+				// Use "old-style" FP instructions. We know that both operands are in memory, so we
+				// can simply emit FILD followed by FSTP
+				to->putByte(0xDF);
+				modRm(to, 5, src);
+				fstpOut(to, instr);
+			} else {
+				// TODO: Need to clear target register first? (use PXOR, 66 0f ef XX)
+				Operand dst = instr->dest();
+
+				// Use CVTSI2SS or CVTSI2SD
+				if (dst.size() == Size::sDouble)
+					to->putByte(0xF2);
+				else
+					to->putByte(0xF3);
+				to->putByte(0x0F);
+				to->putByte(0x2A);
+				modRm(to, fpRegisterId(dst.reg()), src);
+			}
 		}
 
-		void fsubpOut(Output *to, Instr *instr) {
-			to->putByte(0xDE);
-			to->putByte(0xE9);
-		}
+		void ucastfOut(Output *to, Instr *instr) {
+			Operand src = instr->src();
+			Operand dst = instr->dest();
+			if (src.size() == Size::sLong) {
+				// This case is a bit tricky. We need to check the sign bit manually and adjust.
+				// The idea here is to load the number as usual. If it was treated as negative, we
+				// can simply compensate by adding a large enough number to it.
 
-		void fmulpOut(Output *to, Instr *instr) {
-			to->putByte(0xDE);
-			to->putByte(0xC9);
-		}
+				// We emit the following code:
+				// fild <src>
+				// mov <tmp>, <src+4>
+				// test <tmp>, <tmp>
+				// jns DONE
+				// xor <tmp>, <tmp>
+				// mov <dst>, <tmp>
+				// mov <tmp>, #exp
+				// mov <dst+4>, <tmp>
+				// fild <tmp>
+				// faddp ST1
+				// DONE
+				// fstp
 
-		void fdivpOut(Output *to, Instr *instr) {
-			to->putByte(0xDE);
-			to->putByte(0xF9);
-		}
+				// Here, #exp is the number 2^62 as encoded for a double:
+				Nat exp = (64 + 1023) << 20;
 
-		void fcomppOut(Output *to, Instr *instr) {
-			// fcomip ST1
-			to->putByte(0xDF);
-			to->putByte(0xF0 + 1);
+				Reg tmp = asSize(dst.reg(), Size::sInt);
+				Operand stackPos = longRel(ptrStack, Offset());
+				Operand stackLow = low32(stackPos);
+				Operand stackHigh = high32(stackPos);
 
-			// fstp ST0 (effectively only a pop)
-			to->putByte(0xDD);
-			to->putByte(0xD8 + 0);
-		}
-
-		void fwaitOut(Output *to, Instr *instr) {
-			to->putByte(0x9B);
+				// fild <src>
+				to->putByte(0xDF);
+				modRm(to, 5, instr->src());
+				// mov <tmp>, <src+4>
+				movOut(to, tmp, stackHigh);
+				// test <tmp>, <tmp>
+				to->putByte(0x85);
+				modRm(to, registerId(tmp), tmp);
+				// jns DONE
+				to->putByte(0x0F);
+				to->putByte(0x89);
+				to->putInt(20);
+				// xor <tmp>, <tmp>
+				bxorOut(to, tmp, tmp);
+				// mov <dst>, <tmp>
+				movOut(to, stackLow, tmp);
+				// mov <tmp>, #exp
+				movOut(to, tmp, natConst(exp));
+				// mov <dst+4>, <tmp>
+				movOut(to, stackHigh, tmp);
+				// fild <tmp>
+				to->putByte(0xDD);
+				modRm(to, 0, stackPos);
+				// faddp ST0
+				to->putByte(0xDE);
+				to->putByte(0xC1);
+				// fstp <dest>
+				if (dst.size() == Size::sDouble)
+					to->putByte(0xDD);
+				else
+					to->putByte(0xD9);
+				modRm(to, 3, stackPos);
+			} else {
+				// Issue a 64-bit load. Pre-processing ensures that this is fine to do.
+				to->putByte(0xDF);
+				modRm(to, 5, instr->src());
+				fstpOut(to, instr);
+			}
 		}
 
 		void threadLocalOut(Output *to, Instr *instr) {
@@ -641,7 +966,16 @@ namespace code {
 			// We don't currently offer the ability to lookup source locations in the final binary.
 		}
 
+		void metaOut(Output *, Instr *) {
+			// We don't output metadata.
+		}
+
+#define OUTPUT(x) { op::x, &x ## Out }
+
+		typedef void (*OutputFn)(Output *to, Instr *instr);
+
 		const OpEntry<OutputFn> outputMap[] = {
+			OUTPUT(epilog),
 			OUTPUT(mov),
 			OUTPUT(swap),
 			OUTPUT(add),
@@ -653,6 +987,7 @@ namespace code {
 			OUTPUT(sbb),
 			OUTPUT(bxor),
 			OUTPUT(cmp),
+			OUTPUT(test),
 			OUTPUT(push),
 			OUTPUT(pop),
 			OUTPUT(pushFlags),
@@ -672,22 +1007,28 @@ namespace code {
 			OUTPUT(lea),
 
 			OUTPUT(fstp),
-			OUTPUT(fistp),
 			OUTPUT(fld),
-			OUTPUT(fild),
-			OUTPUT(fldz),
-			OUTPUT(faddp),
-			OUTPUT(fsubp),
-			OUTPUT(fmulp),
-			OUTPUT(fdivp),
-			OUTPUT(fcompp),
-			OUTPUT(fwait),
+			OUTPUT(fadd),
+			OUTPUT(fsub),
+			OUTPUT(fmul),
+			OUTPUT(fdiv),
+			OUTPUT(fneg),
+			OUTPUT(fcmp),
+			OUTPUT(fcast),
+			OUTPUT(fcasti),
+			OUTPUT(fcastu),
+			OUTPUT(icastf),
+			OUTPUT(ucastf),
+			// Basic testing shows that:
+			// use movups instead of movaps for unaligned loads
+			// unaligned operations for add/sub/movss/etc is fine
 
 			OUTPUT(threadLocal),
 			OUTPUT(dat),
 			OUTPUT(lblOffset),
 			OUTPUT(align),
 			OUTPUT(location),
+			OUTPUT(meta),
 		};
 
 		void output(Listing *src, Output *to) {
