@@ -10,6 +10,7 @@ namespace code {
 
 #define IMM_REG(x) { op::x, &RemoveInvalid::immRegTfm }
 #define TRANSFORM(x) { op::x, &RemoveInvalid::x ## Tfm }
+#define FP_OP(x) { op::x, &RemoveInvalid::fpInstrTfm }
 
 		const OpEntry<RemoveInvalid::TransformFn> RemoveInvalid::transformMap[] = {
 			IMM_REG(mov),
@@ -21,6 +22,9 @@ namespace code {
 			IMM_REG(sbb),
 			IMM_REG(bxor),
 			IMM_REG(cmp),
+			IMM_REG(test),
+
+			TRANSFORM(beginBlock),
 
 			TRANSFORM(lea),
 			TRANSFORM(mul),
@@ -39,6 +43,18 @@ namespace code {
 			TRANSFORM(fnParamRef),
 			TRANSFORM(fnCall),
 			TRANSFORM(fnCallRef),
+
+			FP_OP(fadd),
+			FP_OP(fsub),
+			TRANSFORM(fneg),
+			FP_OP(fmul),
+			FP_OP(fdiv),
+			FP_OP(fcmp),
+			FP_OP(fcast),
+			TRANSFORM(fcasti),
+			TRANSFORM(fcastu),
+			TRANSFORM(icastf),
+			TRANSFORM(ucastf),
 		};
 
 		RemoveInvalid::Param::Param(Operand src, TypeDesc *type, Bool ref) : src(src), type(type), ref(ref) {}
@@ -113,6 +129,13 @@ namespace code {
 			}
 		}
 
+		void RemoveInvalid::beginBlockTfm(Listing *dest, Instr *instr, Nat line) {
+			// We need to tell the next step what register(s) are free.
+			Reg r = unusedReg(line);
+			if (r != noReg)
+				*dest << instr->alterDest(r);
+		}
+
 		void RemoveInvalid::leaTfm(Listing *dest, Instr *instr, Nat line) {
 
 			// We can encode writing directly to a register.
@@ -677,5 +700,344 @@ namespace code {
 			params->clear();
 		}
 
+		Reg RemoveInvalid::loadFpRegister(Listing *dest, const Operand &op, Nat line) {
+			// Must be in a fp register!
+			if (fpRegister(op))
+				return op.reg();
+
+			// Just load it into a free vector register!
+			Reg r = asSize(unusedFpReg(used->at(line)), op.size());
+			used->at(line)->put(r);
+			*dest << mov(r, op);
+			return r;
+		}
+
+		Operand RemoveInvalid::loadFpRegisterOrMemory(Listing *dest, const Operand &op, Nat line) {
+			switch (op.type()) {
+			case opRelative:
+			case opVariable:
+				return op;
+			default:
+				return loadFpRegister(dest, op, line);
+			}
+		}
+
+		void RemoveInvalid::fpInstrTfm(Listing *dest, Instr *instr, Nat line) {
+			// The XMM instructions we use support a source in memory, but not a destination.
+			Operand dst = instr->dest();
+			DestMode mode = destMode(instr->op());
+
+			Reg dstReg = noReg;
+			if (mode & destRead) {
+				dstReg = loadFpRegister(dest, dst, line);
+			} else {
+				// Just pick a register if the specified one is not good enough.
+				if (fpRegister(dst)) {
+					dstReg = dst.reg();
+				} else {
+					dstReg = asSize(unusedFpReg(used->at(line)), dst.size());
+					// We don't need to update the register in the used set either, usage will not overlap.
+					// No need to load it, it is not read.
+				}
+			}
+
+			Operand src = loadFpRegisterOrMemory(dest, instr->src(), line);
+
+			*dest << instr->alter(dstReg, src);
+
+			// Write it back if necessary.
+			if (mode & destWrite) {
+				if (dst.type() != opRegister || dst.reg() != dstReg) {
+					*dest << mov(dst, dstReg);
+				}
+			}
+		}
+
+		void RemoveInvalid::fnegTfm(Listing *dest, Instr *instr, Nat line) {
+			Operand src = loadFpRegisterOrMemory(dest, instr->src(), line);
+			Operand dst = instr->dest();
+
+			// Just pick a register if the specified one is not good enough.
+			if (!fpRegister(dst)) {
+				Reg dstReg = asSize(unusedFpReg(used->at(line)), dst.size());
+				*dest << instr->alter(dstReg, src);
+				*dest << mov(dst, dstReg);
+			} else {
+				*dest << instr->alterSrc(src);
+			}
+		}
+
+		void RemoveInvalid::fcastiTfm(Listing *dest, Instr *instr, Nat line) {
+			// If the output size is 32-bit we use SSE instructions directly. If 64-bit, it is
+			// easier to fall back to the "old" FP unit as it has an operation for integer
+			// conversions directly.
+			Operand dst = instr->dest();
+			if (dst.size() == Size::sLong) {
+				// x87 FP stack. Source and destination need to be in memory.
+				Operand src = instr->src();
+				bool spillToStack = dst.type() == opRegister || src.type() == opRegister;
+
+				if (spillToStack) {
+					*dest << sub(ptrStack, ptrConst(Size::sLong));
+				}
+
+				if (src.type() == opRegister) {
+					Operand stackTmp = xRel(src.size(), ptrStack, Offset());
+					*dest << mov(stackTmp, src);
+					src = stackTmp;
+				}
+
+				if (dst.type() == opRegister) {
+					*dest << instr->alter(xRel(dst.size(), ptrStack, Offset()), src);
+				} else {
+					*dest << instr->alterSrc(src);
+				}
+
+				if (spillToStack) {
+					if (dst.type() == opRegister) {
+						*dest << pop(low32(dst));
+						*dest << pop(high32(dst));
+					} else {
+						*dest << add(ptrStack, ptrConst(Size::sLong));
+					}
+				}
+
+			} else {
+				// SSE. Need source in fp register or memory. Destination has to be integer register.
+				Operand src = loadFpRegisterOrMemory(dest, instr->src(), line);
+				if (dst.type() != opRegister) {
+					Reg r = asSize(unusedReg(line), dst.size());
+					*dest << instr->alter(r, src);
+					*dest << mov(dst, r);
+				} else {
+					*dest << instr->alterSrc(src);
+				}
+			}
+		}
+
+		void RemoveInvalid::fcastuTfm(Listing *dest, Instr *instr, Nat line) {
+			// In case we are asked to convert to a 32-bit uint, we always use the x87 FP stack.
+			// For the 32-bit output case, we simply convert to a 64-bit int and return the lower
+			// 32 bits. For the 64-bit case, we need to emit a bit more elaborate code that needs
+			// a register and 64-bits of temporary storage on the stack.
+
+			Operand src = instr->src();
+			Operand dst = instr->dest();
+			if (dst.size() == Size::sLong) {
+				Size stackSize = Size::sLong;
+
+				// A temporary register is needed as destination.
+				Reg dstReg = noReg;
+				if (dst.type() == opRegister) {
+					dstReg = dst.reg();
+				}
+				if (dstReg == noReg) {
+					const Reg options[3] = { rax, rbx, rcx };
+					RegSet *used = this->used->at(line);
+					for (Nat i = 0; i < ARRAY_COUNT(options); i++) {
+						if (!used->has(low32(options[i])) && !used->has(high32(options[i]))) {
+							dstReg = options[i];
+							break;
+						}
+					}
+				}
+				Bool saveRax = false;
+				if (dstReg == noReg) {
+					dstReg = rax;
+					saveRax = true;
+					stackSize += Size::sLong;
+				}
+
+				*dest << sub(ptrStack, ptrConst(stackSize));
+
+				if (saveRax) {
+					*dest << mov(intRel(ptrStack, Offset::sLong), eax);
+					*dest << mov(intRel(ptrStack, Offset::sLong + Offset::sInt), edx);
+				}
+
+				// Source needs to be in memory.
+				if (src.type() == opRegister) {
+					Operand stackTmp = xRel(src.size(), ptrStack, Offset());
+					*dest << mov(stackTmp, src);
+					src = stackTmp;
+				}
+
+				*dest << instr->alter(dstReg, src);
+
+				if (saveRax) {
+					*dest << mov(eax, intRel(ptrStack, Offset::sLong));
+					*dest << mov(edx, intRel(ptrStack, Offset::sLong + Offset::sInt));
+				}
+
+				// Copy result to proper location.
+				if (dst.type() != opRegister || dst.reg() != dstReg) {
+					*dest << mov(low32(dst), low32(dstReg));
+					*dest << mov(high32(dst), high32(dstReg));
+				}
+				*dest << add(ptrStack, ptrConst(stackSize));
+			} else {
+				// We always need to spill to the stack since the output will be longer than the target.
+				*dest << sub(ptrStack, ptrConst(Size::sLong));
+
+				if (src.type() == opRegister) {
+					Operand stackTmp = xRel(src.size(), ptrStack, Offset());
+					*dest << mov(stackTmp, src);
+					src = stackTmp;
+				}
+
+				*dest << instr->alter(intRel(ptrStack, Offset()), src);
+
+				if (dst.type() == opRegister) {
+					*dest << mov(dst, intRel(ptrStack, Offset()));
+				} else {
+					Reg tmpReg = unusedReg(line);
+					if (tmpReg == noReg) {
+						*dest << push(eax);
+						*dest << mov(eax, intRel(ptrStack, Offset()));
+						*dest << mov(dst, eax);
+						*dest << pop(eax);
+					} else {
+						tmpReg = asSize(tmpReg, Size::sInt);
+						*dest << mov(tmpReg, intRel(ptrStack, Offset()));
+						*dest << mov(dst, tmpReg);
+					}
+				}
+
+				*dest << add(ptrStack, ptrConst(Size::sLong));
+			}
+		}
+
+		void RemoveInvalid::icastfTfm(Listing *dest, Instr *instr, Nat line) {
+			// If the output size is 32-bit we use SSE instructions directly. If 64-bit, it is
+			// easier to fall back to the "old" FP unit as it has an operation for integer
+			// conversions directly.
+			Operand src = instr->src();
+			if (src.size() == Size::sLong) {
+				// x87 FP stack. Source and destination need to be in memory.
+				Operand dst = instr->dest();
+				bool spillToStack = dst.type() == opRegister || src.type() == opRegister;
+
+				if (spillToStack) {
+					if (src.type() == opRegister) {
+						*dest << push(high32(src));
+						*dest << push(low32(src));
+					} else {
+						*dest << sub(ptrStack, ptrConst(Size::sLong));
+					}
+				}
+
+				if (dst.type() == opRegister) {
+					Operand stackTmp = xRel(dst.size(), ptrStack, Offset());
+					*dest << instr->alter(stackTmp, src);
+					*dest << mov(dst, stackTmp);
+				} else {
+					*dest << instr->alterSrc(src);
+				}
+
+				if (spillToStack) {
+					*dest << add(ptrStack, ptrConst(Size::sLong));
+				}
+
+			} else {
+				// SSE. Need source in register or memory. Destination has to be integer register.
+				Operand dst = instr->dest();
+				if (fpRegister(dst)) {
+					*dest << instr;
+				} else {
+					Reg r = asSize(unusedFpReg(used->at(line)), dst.size());
+					*dest << instr->alterDest(r);
+					*dest << mov(dst, r);
+				}
+			}
+		}
+
+		void RemoveInvalid::ucastfTfm(Listing *dest, Instr *instr, Nat line) {
+			// For unsigned values, we always use the x87 stack. For Nat, we simply spill it to the
+			// stack and load a 64-bit integer value. For Word, we need slightly more complex
+			// machine code. That code requires the source to be in a register, the destination in
+			// memory, and 8 bytes of stack space allocated for it.
+
+			Operand src = instr->src();
+			Operand dst = instr->dest();
+			Size stackSize = Size::sLong;
+
+			// We need a temporary register. If dest is a register, we can use that.
+			Reg tmpReg = noReg;
+			if (dst.type() == opRegister)
+				tmpReg = dst.reg();
+			if (tmpReg == noReg)
+				tmpReg = unusedReg(line);
+			Bool saveEax = false;
+			if (tmpReg == noReg) {
+				tmpReg = eax;
+				saveEax = true;
+				stackSize += Size::sLong;
+			}
+
+			tmpReg = asSize(tmpReg, Size::sInt);
+
+			*dest << sub(ptrStack, ptrConst(stackSize));
+
+			if (saveEax)
+				*dest << mov(intRel(ptrStack, Offset::sLong), eax);
+
+			Bool copyResult = false;
+
+
+			if (src.size() == Size::sLong) {
+				// Source needs to be in memory.
+				if (src.type() == opRegister) {
+					*dest << mov(intRel(ptrStack, Offset()), low32(src));
+					*dest << mov(intRel(ptrStack, Offset::sInt), high32(src));
+				} else {
+					*dest << mov(tmpReg, low32(src));
+					*dest << mov(intRel(ptrStack, Offset()), tmpReg);
+					*dest << mov(tmpReg, high32(src));
+					*dest << mov(intRel(ptrStack, Offset::sInt), tmpReg);
+				}
+
+				// Inform the next step where the result is located.
+				*dest << instr->alter(asSize(tmpReg, dst.size()), longRel(ptrStack, Offset()));
+				copyResult = true;
+
+			} else {
+				// We always need to copy to the stack to ensure that we have space for zero
+				// extension.
+
+				if (src.type() == opRegister) {
+					*dest << mov(intRel(ptrStack, Offset()), src);
+				} else {
+					*dest << mov(tmpReg, src);
+					*dest << mov(intRel(ptrStack, Offset()), tmpReg);
+				}
+				*dest << bxor(tmpReg, tmpReg);
+				*dest << mov(intRel(ptrStack, Offset::sInt), tmpReg);
+
+				if (dst.type() == opRegister) {
+					*dest << instr->alter(xRel(dst.size(), ptrStack, Offset()), intRel(ptrStack, Offset()));
+					copyResult = true;
+				} else {
+					*dest << instr->alterSrc(intRel(ptrStack, Offset()));
+				}
+			}
+
+
+			if (saveEax)
+				*dest << mov(eax, intRel(ptrStack, Offset::sLong));
+
+			// Copy result back to where it belongs.
+			if (copyResult) {
+				if (dst.type() == opRegister) {
+					*dest << mov(dst, xRel(dst.size(), ptrStack, Offset()));
+				} else {
+					Reg fpReg = asSize(unusedFpReg(used->at(line)), dst.size());
+					*dest << mov(fpReg, xRel(dst.size(), ptrStack, Offset()));
+					*dest << mov(dst, fpReg);
+				}
+			}
+
+			*dest << add(ptrStack, ptrConst(Size::sLong));
+		}
+
 	}
 }
