@@ -42,6 +42,89 @@ namespace storm {
 	};
 
 
+	/**
+	 * Runtime errors from various parts of the system.
+	 */
+	class EXCEPTION_EXPORT RuntimeError : public Exception {
+		STORM_EXCEPTION;
+	public:
+		STORM_CTOR RuntimeError();
+	};
+
+
+	/**
+	 * Exception thrown by the garbage collector.
+	 *
+	 * The most common type of error is an out of memory error, but others may occur.
+	 */
+	class EXCEPTION_EXPORT GcError : public RuntimeError {
+		STORM_EXCEPTION;
+	public:
+		// Create, basic C-string to avoid memory allocations.
+		GcError(const wchar *msg);
+
+		virtual void STORM_FN message(StrBuf *to) const;
+
+	private:
+		// The message.
+		const wchar *msg;
+	};
+
+
+	/**
+	 * Arithmetic errors. Thrown by numerical operations when an error occurs.
+	 */
+	class EXCEPTION_EXPORT NumericError : public RuntimeError {
+		STORM_EXCEPTION;
+	public:
+		// Create.
+		STORM_CTOR NumericError();
+	};
+
+
+	/**
+	 * Division by zero. Thrown on integer divisions by zero. Floating point numbers generate
+	 * infinity/nan in many cases instead.
+	 */
+	class EXCEPTION_EXPORT DivisionByZero : public NumericError {
+		STORM_EXCEPTION;
+	public:
+		// Create.
+		STORM_CTOR DivisionByZero();
+
+		// Message.
+		virtual void STORM_FN message(StrBuf *to) const;
+	};
+
+
+	/**
+	 * Access violation. Thrown when an invalid address has been accessed.
+	 */
+	class EXCEPTION_EXPORT MemoryAccessError : public RuntimeError {
+		STORM_EXCEPTION;
+	public:
+		enum Type {
+			notMapped,
+			invalidAccess,
+			invalidAlignment,
+			kernel,
+		};
+
+		// Create. If 'mapped' is true, then the address is mapped, but the access type was invalid.
+		STORM_CTOR MemoryAccessError(Word address, Type type);
+
+		// Message.
+		virtual void STORM_FN message(StrBuf *to) const;
+
+	private:
+		// Faulting address.
+		Word address;
+
+		// Type of fault.
+		Type type;
+	};
+
+
 	/**
 	 * Generic exceptions.
 	 */
@@ -58,10 +141,26 @@ namespace storm {
 	};
 
 
+	/**
+	 * Invalid usage.
+	 */
+	class EXCEPTION_EXPORT UsageError : public Exception {
+		STORM_EXCEPTION;
+	public:
+		UsageError(const wchar *msg);
+		STORM_CTOR UsageError(Str *msg);
+
+		virtual void STORM_FN message(StrBuf *to) const;
+
+	private:
+		Str *msg;
+	};
+
+
 	/**
 	 * Internal error.
 	 */
-	class EXCEPTION_EXPORT InternalError : public Exception {
+	class EXCEPTION_EXPORT InternalError : public RuntimeError {
 		STORM_EXCEPTION;
 	public:
 		InternalError(const wchar *msg);
@@ -73,29 +172,19 @@ namespace storm {
 	};
 
 
-	/**
-	 * Runtime errors.
-	 */
-	class EXCEPTION_EXPORT RuntimeError : public Exception {
-		STORM_EXCEPTION;
-	public:
-		RuntimeError(const wchar *msg);
-		STORM_CTOR RuntimeError(Str *msg);
-
-		virtual void STORM_FN message(StrBuf *to) const;
-	private:
-		Str *msg;
-	};
-
-
 	/**
 	 * Calling an abstract function.
 	 */
 	class EXCEPTION_EXPORT AbstractFnCalled : public RuntimeError {
+		STORM_EXCEPTION;
 	public:
 		AbstractFnCalled(const wchar *name);
 		STORM_CTOR AbstractFnCalled(Str *msg);
+
 		virtual void STORM_FN message(StrBuf *to) const;
+
+	private:
+		Str *name;
 	};
 
 
