@@ -19,7 +19,7 @@ namespace gui {
 
 	Window::Window() :
 		myHandle(invalid), myParent(null), myRoot(null),
-		myVisible(true), drawing(false), mouseInside(false),
+		myVisible(true), myEnabled(true), drawing(false), mouseInside(false),
 		gdkWindow(null), gTimer(null),
 		myPos(0, 0, 40, 40) /* large enought to not generate warnings in Gtk+ */ {
 
@@ -146,6 +146,10 @@ namespace gui {
 		return myVisible;
 	}
 
+	Bool Window::enabled() {
+		return myEnabled;
+	}
+
 	Font *Window::font() {
 		return new (this) Font(*myFont);
 	}
@@ -219,6 +223,7 @@ namespace gui {
 				return msgResult(0);
 			break;
 		case WM_LBUTTONDBLCLK:
+			onClick(true, dpiFromPx(currentDpi(), mousePos(msg)), mouse::left);
 			if (onDblClick(dpiFromPx(currentDpi(), mousePos(msg)), mouse::left))
 				return msgResult(0);
 			break;
@@ -231,6 +236,7 @@ namespace gui {
 				return msgResult(0);
 			break;
 		case WM_MBUTTONDBLCLK:
+			onClick(true, dpiFromPx(currentDpi(), mousePos(msg)), mouse::middle);
 			if (onDblClick(dpiFromPx(currentDpi(), mousePos(msg)), mouse::middle))
 				return msgResult(0);
 			break;
@@ -243,6 +249,7 @@ namespace gui {
 				return msgResult(0);
 			break;
 		case WM_RBUTTONDBLCLK:
+			onClick(true, dpiFromPx(currentDpi(), mousePos(msg)), mouse::right);
 			if (onDblClick(dpiFromPx(currentDpi(), mousePos(msg)), mouse::right))
 				return msgResult(0);
 			break;
@@ -304,6 +311,12 @@ namespace gui {
 			ShowWindow(handle().hwnd(), show ? TRUE : FALSE);
 	}
 
+	void Window::enabled(Bool enable) {
+		myEnabled = enable;
+		if (created())
+			EnableWindow(handle().hwnd(), enable ? TRUE : FALSE);
+	}
+
 	Str *Window::text() {
 		if (created()) {
 			Nat len = GetWindowTextLength(handle().hwnd());
@@ -452,6 +465,9 @@ namespace gui {
 			}
 		}
 
+		if (!myEnabled)
+			style |= WS_DISABLED;
+
 		Point p = myPos.p0;
 		Size s = myPos.size();
 
@@ -611,6 +627,9 @@ namespace gui {
 		else
 			gtk_widget_hide(widget);
 
+		if (!myEnabled)
+			gtk_widget_set_sensitive(widget, FALSE);
+
 		if (myFont != app(engine())->defaultFont)
 			gtk_widget_override_font(fontWidget(), myFont->desc());
 
@@ -661,6 +680,8 @@ namespace gui {
 		if (!widget)
 			return;
 
+		// Note: This might not be entirely correct in nested structures.
+
 		for (GtkWidget *at = widget; !gtk_widget_get_has_window(at); at = gtk_widget_get_parent(at)) {
 			GtkAllocation delta;
 			gtk_widget_get_allocation(at, &delta);
@@ -671,22 +692,43 @@ namespace gui {
 	}
 
 	// Translate window coordinates to widget coordinates.
-	static bool translatePoint(GtkWidget *widget, gdouble x, gdouble y, Point &out) {
+	static bool translatePoint(GtkWidget *eventWidget, GtkWidget *to, gdouble x, gdouble y, Point &out) {
 		GtkAllocation alloc;
-		gtk_widget_get_allocation(widget, &alloc);
+		gtk_widget_get_allocation(to, &alloc);
 
-		// Traverse upwards until we find a widget with a window. That's where the event originated from!
-		for (GtkWidget *at = widget; !gtk_widget_get_has_window(at); at = gtk_widget_get_parent(at)) {
-			GtkAllocation delta;
-			gtk_widget_get_allocation(at, &delta);
+		// Make relative to 'to's window.
+		x -= alloc.x;
+		y -= alloc.y;
 
-			x -= delta.x;
-			y -= delta.y;
+		// Check if there are any other windows on the way.
+		for (GtkWidget *at = gtk_widget_get_parent(to);
+			 at != null && at != eventWidget;
+			 at = gtk_widget_get_parent(at)) {
+
+			if (gtk_widget_get_has_window(at)) {
+				GtkAllocation diff;
+				gtk_widget_get_allocation(at, &diff);
+
+				// New window, update.
+				x -= diff.x;
+				x -= diff.y;
+			}
+		}
+
+		// If 'eventWidget' does not have a window, we need to update coordinates.
+		if (eventWidget && !gtk_widget_get_has_window(eventWidget)) {
+			GtkAllocation diff;
+			gtk_widget_get_allocation(eventWidget, &diff);
+
+			x += diff.x;
+			y += diff.y;
 		}
 
 		out.x = x;
 		out.y = y;
 
+		// PLN(L"Result: " << out.x << ", " << out.y);
+
 		return out.x >= 0.0f && out.y >= 0.0f
 			&& out.x < alloc.width && out.y < alloc.height;
 	}
@@ -694,7 +736,7 @@ namespace gui {
 	gboolean Window::onButton(GdkEvent *event) {
 		Point pt;
 		GdkEventButton &b = event->button;
-		if (!translatePoint(drawWidget(), b.x, b.y, pt))
+		if (!translatePoint(handle().widget(), drawWidget(), b.x, b.y, pt))
 			return FALSE;
 
 		mouse::MouseButton button = mouse::MouseButton(b.button - 1);
@@ -719,7 +761,7 @@ namespace gui {
 	gboolean Window::onMotion(GdkEvent *event) {
 		Point pt;
 		GdkEventMotion &m = event->motion;
-		if (!translatePoint(drawWidget(), m.x, m.y, pt))
+		if (!translatePoint(handle().widget(), drawWidget(), m.x, m.y, pt))
 			return FALSE;
 
 		return onMouseMove(pt) ? TRUE : FALSE;
@@ -742,7 +784,7 @@ namespace gui {
 	gboolean Window::onScroll(GdkEvent *event) {
 		Point pt;
 		GdkEventScroll &s = event->scroll;
-		if (!translatePoint(drawWidget(), s.x, s.y, pt))
+		if (!translatePoint(handle().widget(), drawWidget(), s.x, s.y, pt))
 			return FALSE;
 
 		gdouble dx = 0, dy = 0;
@@ -788,16 +830,19 @@ namespace gui {
 		Size s(drawAlloc.width, drawAlloc.height);
 		myPos.size(s);
 
-		if (myPainter)
-			myPainter->uiResize(s, 1);
+		if (myPainter) {
+			int scale = gtk_widget_get_scale_factor(draw);
+			myPainter->uiResize(s * scale, scale);
+		}
 		onResize(s);
 
 		// Queue a repaint for this window. Otherwise, resizing windows that partially contain some
 		// continuous rendering will not always be updated properly. We're probably in a race with
 		// Gtk when we just update a part of the window.
 		GdkWindow *window = gtk_widget_get_window(handle().widget());
-		if (window)
+		if (window) {
 			gdk_window_invalidate_rect(window, NULL, false);
+		}
 	}
 
 	bool Window::preExpose(GtkWidget *widget) {
@@ -1110,6 +1155,13 @@ namespace gui {
 		}
 	}
 
+	void Window::enabled(Bool v) {
+		myEnabled = v;
+		if (created()) {
+			gtk_widget_set_sensitive(handle().widget(), v ? TRUE : FALSE);
+		}
+	}
+
 	void Window::focus() {
 		if (created())
 			gtk_widget_grab_focus(handle().widget());
