@@ -6,7 +6,6 @@
 #include "Sync.h"
 #include "Utils/Bitwise.h"
 #include "Utils/Lock.h"
-#include "UThreadX64.h"
 #include <limits>
 
 #ifdef POSIX
@@ -601,11 +600,12 @@ namespace os {
 		UThreadState::current()->reap();
 	}
 
+
 	/**
-	 * Machine specific code.
+	 * OS-specific code.
 	 */
 
-#if defined(X86)
+#if defined(WINDOWS)
 
 	static int64 timestamp() {
 		LARGE_INTEGER v;
@@ -629,6 +629,40 @@ namespace os {
 		return nat((1000 * (target - now)) / v.QuadPart);
 	}
 
+#elif defined(POSIX)
+
+	static int64 timestamp() {
+		struct timespec time = {0, 0};
+		clock_gettime(CLOCK_MONOTONIC, &time);
+
+		int64 r = time.tv_sec;
+		r *= 1000 * 1000;
+		r += time.tv_nsec / 1000;
+		return r;
+	}
+
+	static int64 msInTimestamp(nat ms) {
+		return int64(ms * 1000);
+	}
+
+	static nat remainingMs(int64 target) {
+		int64 now = timestamp();
+		if (target <= now)
+			return 0;
+
+		int64 remaining = target - now;
+		return nat(remaining / 1000);
+	}
+
+#endif
+
+
+	/**
+	 * Machine specific code.
+	 */
+
+#if defined(X86)
+
 	static void *&stackPtr(Stack &stack) {
 		return stack.desc->low;
 	}
@@ -659,7 +693,7 @@ namespace os {
 		push(0); // edi
 		push(-1); // seh (end of list is -1)
 		push(stack.desc->high);
-		push(stack.base());
+		push(stack.limit());
 		push(stack.desc->low); // current stack pointer (approximate, this is enough)
 
 		// Set the 'desc' of 'stack' to the actual stack pointer, so that we can run code on this stack.
@@ -796,29 +830,6 @@ namespace os {
 
 #elif defined(GCC) && defined(X64)
 
-	static int64 timestamp() {
-		struct timespec time = {0, 0};
-		clock_gettime(CLOCK_MONOTONIC, &time);
-
-		int64 r = time.tv_sec;
-		r *= 1000 * 1000;
-		r += time.tv_nsec / 1000;
-		return r;
-	}
-
-	static int64 msInTimestamp(nat ms) {
-		return int64(ms * 1000);
-	}
-
-	static nat remainingMs(int64 target) {
-		int64 now = timestamp();
-		if (target <= now)
-			return 0;
-
-		int64 remaining = target - now;
-		return nat(remaining / 1000);
-	}
-
 	static void *&stackPtr(Stack &stack) {
 		return stack.desc->low;
 	}
@@ -919,7 +930,7 @@ namespace os {
 	}
 
 	void UThreadData::restoreContext(StackDesc *desc) {
-		atomicWrite(stack.desc, (StackDesc *)(desc));
+		atomicWrite(stack.desc, (StackDesc *)desc);
 	}
 
 	// Called from UThreadX64.S
@@ -936,6 +947,120 @@ namespace os {
 			: address(doEndDetour);
 	}
 
+#elif defined(GCC) && defined(ARM64)
+
+	static void *&stackPtr(Stack &stack) {
+		return stack.desc->low;
+	}
+
+	void UThreadData::push(void *v) {
+		void **sp = (void **)stackPtr(stack);
+		*--sp = v;
+		stackPtr(stack) = sp;
+	}
+
+	void UThreadData::push(uintptr_t v) {
+		void **sp = (void **)stackPtr(stack);
+		*--sp = (void *)v;
+		stackPtr(stack) = sp;
+	}
+
+	void UThreadData::push(intptr_t v) {
+		void **sp = (void **)stackPtr(stack);
+		*--sp = (void *)v;
+		stackPtr(stack) = sp;
+	}
+
+	void UThreadData::pushContext(const void *fn) {
+		pushContext(fn, null);
+	}
+
+	// Function for launching new threads. Not callable from C, so a regular variable.
+	extern "C" char launchThreadStub;
+
+	const int arm64FrameWords = 24; // Size of the arm64 context in 8-byte words.
+
+	void UThreadData::pushContext(const void *fn, void *param) {
+		void **sp = (void **)stackPtr(stack);
+
+		// We need a bit of padding for exceptions to work. They read a few words here (maybe due to
+		// incorrect unwind info?) Remember: Stack must be 16-byte aligned.
+		// sp -= 2*sizeof(void *);
+
+		sp -= arm64FrameWords;
+		stackPtr(stack) = sp;
+
+		// Zero everything.
+		memset(sp, 0, sizeof(void *)*arm64FrameWords);
+
+		// Store the state we need.
+		sp[1] = &launchThreadStub; // Return to.
+		sp[23] = param;
+		sp[20] = stack.desc->low;
+		sp[21] = stack.desc->dummy;
+		sp[22] = stack.desc->high;
+
+		// Parameters to 'launchThread':
+		sp[0] = null; // base pointer of prev frame.
+		sp[2] = (void *)fn; // function to call
+		sp[3] = param; // parameter
+		sp[4] = null; // apparent caller (none).
+
+		// Now we can use the new Desc on the stack.
+		atomicWrite(stack.desc, (StackDesc *)(sp + 20));
+	}
+
+	// Logical location where thread switches are made. Used to get good backtraces.
+	// Note: This is a label in asm, so the interesting thing about this is its address.
+	extern "C" char doSwitchReturnLoc;
+
+	StackDesc *UThreadData::pushSubContext(const void *fn, void *param) {
+		void **sp = (void **)stackPtr(stack);
+		void *oldSp = sp;
+		sp -= arm64FrameWords;
+		// Note: We don't save the sp in the old desc, since we don't want to trash it!
+
+		// Zero the stack frame.
+		memset(sp, 0, sizeof(void *)*arm64FrameWords);
+
+		// Store the state we need. Similar to pushContext.
+		sp[1] = &launchThreadStub; // Return to.
+		sp[20] = sp; // low of new desc.
+		sp[21] = stack.desc->dummy;
+		sp[22] = stack.desc->high;
+
+		// Parameters to 'launchThread':
+		sp[0] = oldSp; // stored frame pointer.
+		sp[2] = (void *)fn; // function to call.
+		sp[3] = param; // parameter
+		sp[4] = &doSwitchReturnLoc; // apparent return address (for stack traces to work properly)
+
+		StackDesc *old = stack.desc;
+
+		// Set the new desc.
+		atomicWrite(stack.desc, (StackDesc *)(sp + 20));
+
+		return old;
+	}
+
+	void UThreadData::restoreContext(StackDesc *desc) {
+		atomicWrite(stack.desc, (StackDesc *)desc);
+	}
+
+	// Called from ASM.
+	extern "C" void doEndDetour2(void **result) {
+		UThreadState::current()->endDetour(result);
+	}
+
+	extern "C" void doEndDetour();
+	extern "C" void doEndDetourMember();
+
+	static const void *endDetourFn(bool member) {
+		return member
+			? address(doEndDetourMember)
+			: address(doEndDetour);
+	}
+
 #endif
 
 }
