@@ -1,9 +1,19 @@
 #include "stdafx.h"
+#include "CallCommon.h"
 #include "Code/Binary.h"
 #include "Code/Listing.h"
 #include "Code/X64/Arena.h"
 #include "Code/X64/Asm.h"
+#include "Code/Arm64/Arena.h"
+#include "Code/Arm64/Asm.h"
 #include "Compiler/Debug.h"
+#include "Core/Geometry/Rect.h"
+
+/**
+ * File containing tests for calling conventions from the caller's perspective.
+ *
+ * Mirrors tests in Callee.cpp
+ */
 
 using namespace code;
 
@@ -61,8 +71,8 @@ BEGIN_TEST(CallPrimitiveRef, Code) {
 	CHECK_EQ((*fn)(), 102);
 } END_TEST
 
-Int CODECALL callIntManyFn(Int a, Int b, Int c, Int d, Int e, Int f, Int g, Int h) {
-	return a*10000000 + b*1000000 + c*100000 + d*10000 + e*1000 + f*100 + g*10 + h;
+Int CODECALL callIntManyFn(Int a, Int b, Int c, Int d, Int e, Int f, Int g, Int h, Int i) {
+	return a*100000000 + b*10000000 + c*1000000 + d*100000 + e*10000 + f*1000 + g*100 + h*10 + i;
 }
 
 BEGIN_TEST(CallPrimitiveMany, Code) {
@@ -71,7 +81,7 @@ BEGIN_TEST(CallPrimitiveMany, Code) {
 	Engine &e = gEngine();
 	Arena *arena = code::arena(e);
 
-	Ref intFn = arena->external(S("intFn"), address(&callIntManyFn));
+	Ref intFn = arena->external(S("intManyFn"), address(&callIntManyFn));
 
 	Listing *l = new (e) Listing();
 	l->result = intDesc(e);
@@ -79,14 +89,16 @@ BEGIN_TEST(CallPrimitiveMany, Code) {
 
 	*l << mov(eax, intConst(1));
 	*l << mov(ecx, intConst(2));
+	*l << mov(ebx, intConst(3));
 	*l << fnParam(intDesc(e), intConst(3));
 	*l << fnParam(intDesc(e), intConst(4));
 	*l << fnParam(intDesc(e), intConst(5));
 	*l << fnParam(intDesc(e), intConst(6));
 	*l << fnParam(intDesc(e), intConst(7));
 	*l << fnParam(intDesc(e), ecx);
-	*l << fnParam(intDesc(e), intConst(8));
+	*l << fnParam(intDesc(e), intConst(8)); // Stack on X64
 	*l << fnParam(intDesc(e), eax);
+	*l << fnParam(intDesc(e), ebx); // Stack on ARM64
 	*l << fnCall(intFn, false, intDesc(e), ecx);
 
 	*l << fnRet(ecx);
@@ -95,10 +107,10 @@ BEGIN_TEST(CallPrimitiveMany, Code) {
 	typedef Int (*Fn)();
 	Fn fn = (Fn)b->address();
 
-	CHECK_EQ((*fn)(), 34567281);
+	CHECK_EQ((*fn)(), 345672813);
 } END_TEST
 
-#ifdef X64
+#if defined(X64)
 
 // Only usable on X86-64, as we rely on platform specific registers.
 BEGIN_TEST(CallPrimitiveMany64, Code) {
@@ -107,7 +119,7 @@ BEGIN_TEST(CallPrimitiveMany64, Code) {
 	Engine &e = gEngine();
 	Arena *arena = code::arena(e);
 
-	Ref intFn = arena->external(S("intFn"), address(&callIntManyFn));
+	Ref intFn = arena->external(S("intManyFn"), address(&callIntManyFn));
 
 	Listing *l = new (e) Listing();
 	l->result = intDesc(e);
@@ -129,6 +141,7 @@ BEGIN_TEST(CallPrimitiveMany64, Code) {
 	*l << fnParam(intDesc(e), edi);
 	*l << fnParam(intDesc(e), intConst(0));
 	*l << fnParam(intDesc(e), intConst(0));
+	*l << fnParam(intDesc(e), intConst(0));
 	*l << fnCall(intFn, false, intDesc(e), ecx);
 
 	*l << fnRet(ecx);
@@ -137,28 +150,121 @@ BEGIN_TEST(CallPrimitiveMany64, Code) {
 	typedef Int (*Fn)();
 	Fn fn = (Fn)b->address();
 
-	CHECK_EQ((*fn)(), 23456100);
+	CHECK_EQ((*fn)(), 234561000);
+} END_TEST
+
+#elif defined(ARM64)
+
+// Only usable on ARM64, as we rely on platform specific registers.
+BEGIN_TEST(CallPrimitiveMany64, Code) {
+	// Call a function with parameters forming a cycle that the backend needs to break in order to
+	// properly assign the registers.
+	Engine &e = gEngine();
+	Arena *arena = code::arena(e);
+
+	Ref intFn = arena->external(S("intManyFn"), address(&callIntManyFn));
+
+	Listing *l = new (e) Listing();
+	l->result = intDesc(e);
+	*l << prolog();
+
+	using namespace code::arm64;
+	*l << mov(wr(0), intConst(1));
+	*l << mov(wr(1), intConst(2));
+	*l << mov(wr(2), intConst(3));
+	*l << mov(wr(3), intConst(4));
+	*l << mov(wr(4), intConst(5));
+	*l << mov(wr(5), intConst(6));
+	*l << mov(wr(6), intConst(7));
+	*l << mov(wr(7), intConst(8));
+	*l << mov(wr(8), intConst(9));
+
+	*l << fnParam(intDesc(e), wr(1));
+	*l << fnParam(intDesc(e), wr(2));
+	*l << fnParam(intDesc(e), wr(3));
+	*l << fnParam(intDesc(e), wr(4));
+	*l << fnParam(intDesc(e), wr(5));
+	*l << fnParam(intDesc(e), wr(6));
+	*l << fnParam(intDesc(e), wr(7));
+	*l << fnParam(intDesc(e), wr(0));
+	*l << fnParam(intDesc(e), wr(8)); // Passed on the stack.
+	*l << fnCall(intFn, false, intDesc(e), ecx);
+
+	*l << fnRet(ecx);
+
+	Binary *b = new (e) Binary(arena, l);
+	typedef Int (*Fn)();
+	Fn fn = (Fn)b->address();
+
+	CHECK_EQ((*fn)(), 234567819);
 } END_TEST
 
 #endif
 
-struct SmallIntParam {
-	size_t a;
-	size_t b;
-
-	// Make sure it is not a POD.
-	SmallIntParam(size_t a, size_t b) : a(a), b(b) {}
-};
-
-SimpleDesc *smallIntDesc(Engine &e) {
-	SimpleDesc *desc = new (e) SimpleDesc(Size::sPtr * 2, 2);
-	desc->at(0) = Primitive(primitive::integer, Size::sPtr, Offset());
-	desc->at(1) = Primitive(primitive::integer, Size::sPtr, Offset::sPtr);
-	return desc;
+Int CODECALL callTinyInt(TinyIntParam p) {
+	return p.a - p.b;
 }
 
+BEGIN_TEST(CallTinyInt, Code) {
+	Engine &e = gEngine();
+	Arena *arena = code::arena(e);
+
+	Ref intFn = arena->external(S("tinyIntFn"), address(&callTinyInt));
+	SimpleDesc *desc = tinyIntDesc(e);
+
+	Listing *l = new (e) Listing();
+	l->result = intDesc(e);
+	Var v = l->createVar(l->root(), desc->size());
+
+	*l << prolog();
+
+	*l << mov(intRel(v, Offset()), intConst(40));
+	*l << mov(intRel(v, Offset::sInt), intConst(10));
+
+	*l << fnParam(desc, v);
+	*l << fnCall(intFn, false, intDesc(e), eax);
+
+	*l << fnRet(eax);
+
+	Binary *b = new (e) Binary(arena, l);
+	typedef Int (*Fn)();
+	Fn fn = (Fn)b->address();
+
+	CHECK_EQ((*fn)(), 30);
+} END_TEST
+
+BEGIN_TEST(CallTinyIntRef, Code) {
+	Engine &e = gEngine();
+	Arena *arena = code::arena(e);
+
+	Ref intFn = arena->external(S("tinyIntFn"), address(&callTinyInt));
+	SimpleDesc *desc = tinyIntDesc(e);
+
+	Listing *l = new (e) Listing();
+	l->result = intDesc(e);
+	Var v = l->createVar(l->root(), desc->size());
+
+	*l << prolog();
+
+	*l << mov(intRel(v, Offset()), intConst(40));
+	*l << mov(intRel(v, Offset::sInt), intConst(10));
+
+	*l << lea(ptrA, v);
+	*l << fnParamRef(desc, ptrA);
+	*l << fnCall(intFn, false, intDesc(e), eax);
+
+	*l << fnRet(eax);
+
+	Binary *b = new (e) Binary(arena, l);
+	typedef Int (*Fn)();
+	Fn fn = (Fn)b->address();
+
+	CHECK_EQ((*fn)(), 30);
+} END_TEST
+
+
 Int CODECALL callSmallInt(SmallIntParam p) {
-	return Int(p.a + p.b);
+	return Int(p.a - p.b);
 }
 
 BEGIN_TEST(CallSmallInt, Code) {
@@ -174,8 +280,8 @@ BEGIN_TEST(CallSmallInt, Code) {
 
 	*l << prolog();
 
-	*l << mov(ptrRel(v, Offset()), ptrConst(10));
-	*l << mov(ptrRel(v, Offset::sPtr), ptrConst(40));
+	*l << mov(ptrRel(v, Offset()), ptrConst(40));
+	*l << mov(ptrRel(v, Offset::sPtr), ptrConst(10));
 
 	*l << fnParam(desc, v);
 	*l << fnCall(intFn, false, intDesc(e), eax);
@@ -186,7 +292,7 @@ BEGIN_TEST(CallSmallInt, Code) {
 	typedef Int (*Fn)();
 	Fn fn = (Fn)b->address();
 
-	CHECK_EQ((*fn)(), 50);
+	CHECK_EQ((*fn)(), 30);
 } END_TEST
 
 BEGIN_TEST(CallSmallIntRef, Code) {
@@ -202,8 +308,8 @@ BEGIN_TEST(CallSmallIntRef, Code) {
 
 	*l << prolog();
 
-	*l << mov(ptrRel(v, Offset()), ptrConst(10));
-	*l << mov(ptrRel(v, Offset::sPtr), ptrConst(40));
+	*l << mov(ptrRel(v, Offset()), ptrConst(40));
+	*l << mov(ptrRel(v, Offset::sPtr), ptrConst(10));
 
 	*l << lea(ptrA, v);
 	*l << fnParamRef(desc, ptrA);
@@ -215,26 +321,9 @@ BEGIN_TEST(CallSmallIntRef, Code) {
 	typedef Int (*Fn)();
 	Fn fn = (Fn)b->address();
 
-	CHECK_EQ((*fn)(), 50);
+	CHECK_EQ((*fn)(), 30);
 } END_TEST
 
-struct LargeIntParam {
-	size_t a;
-	size_t b;
-	size_t c;
-
-	// Make sure it is not a POD.
-	LargeIntParam(size_t a, size_t b, size_t c) : a(a), b(b), c(c) {}
-};
-
-SimpleDesc *largeIntDesc(Engine &e) {
-	SimpleDesc *desc = new (e) SimpleDesc(Size::sPtr * 3, 3);
-	desc->at(0) = Primitive(primitive::integer, Size::sPtr, Offset());
-	desc->at(1) = Primitive(primitive::integer, Size::sPtr, Offset::sPtr);
-	desc->at(2) = Primitive(primitive::integer, Size::sPtr, Offset::sPtr * 2);
-	return desc;
-}
-
 Int CODECALL callMixedInt(LargeIntParam a, SmallIntParam b) {
 	return Int(a.a + a.b - a.c + b.a - b.b);
 }
@@ -243,7 +332,7 @@ BEGIN_TEST(CallMixedInt, Code) {
 	Engine &e = gEngine();
 	Arena *arena = code::arena(e);
 
-	Ref intFn = arena->external(S("intFn"), address(&callMixedInt));
+	Ref intFn = arena->external(S("mixedIntFn"), address(&callMixedInt));
 	SimpleDesc *small = smallIntDesc(e);
 	SimpleDesc *large = largeIntDesc(e);
 
@@ -270,7 +359,6 @@ BEGIN_TEST(CallMixedInt, Code) {
 	Binary *bin = new (e) Binary(arena, l);
 	typedef Int (*Fn)();
 	Fn fn = (Fn)bin->address();
-
 	CHECK_EQ((*fn)(), 30);
 } END_TEST
 
@@ -312,23 +400,6 @@ BEGIN_TEST(CallMixedIntRef, Code) {
 	CHECK_EQ((*fn)(), 30);
 } END_TEST
 
-struct MixedParam {
-	size_t a;
-	Float b;
-	Float c;
-
-	// Make sure it is not a POD.
-	MixedParam(size_t a, Float b, Float c) : a(a), b(b), c(c) {}
-};
-
-SimpleDesc *mixedDesc(Engine &e) {
-	SimpleDesc *desc = new (e) SimpleDesc(Size::sPtr + Size::sFloat * 2, 3);
-	desc->at(0) = Primitive(primitive::integer, Size::sPtr, Offset());
-	desc->at(1) = Primitive(primitive::real, Size::sFloat, Offset::sPtr);
-	desc->at(2) = Primitive(primitive::real, Size::sFloat, Offset::sPtr + Offset::sFloat);
-	return desc;
-}
-
 Float CODECALL callMixed(MixedParam a) {
 	return Float(a.a) - a.b / a.c;
 }
@@ -484,10 +555,12 @@ BEGIN_TEST(CallFromArray, Code) {
 
 	*l << prolog();
 
-	// Use a 'bad' register for the current backend.
+	// Use a 'bad' register for the current backend. ptrDi on X64, ptrA on ARM, no particular on x86.
 	Reg reg = ptrA;
 	if (as<code::x64::Arena>(arena))
 		reg = code::x64::ptrDi;
+	if (as<code::arm64::Arena>(arena))
+		reg = ptrA;
 
 	*l << mov(reg, params);
 	*l << fnParamRef(valDesc, ptrRel(reg, Offset()));
@@ -509,7 +582,7 @@ BEGIN_TEST(CallFromArray, Code) {
 } END_TEST
 
 static Int CODECALL complexIntFn(DbgVal a, DbgVal b, DbgVal c) {
-	return 100*a.v + 10*b.v + c.v;
+	return 100*a.get() + 10*b.get() + c.get();
 }
 
 BEGIN_TEST(CallComplexFromArray, Code) {
@@ -518,7 +591,7 @@ BEGIN_TEST(CallComplexFromArray, Code) {
 	Type *dbgVal = DbgVal::stormType(e);
 	ComplexDesc *valDesc = new (e) ComplexDesc(dbgVal->size(), dbgVal->copyCtor()->ref(), dbgVal->destructor()->ref());
 
-	Ref toCall = arena->external(S("intFn"), address(&complexIntFn));
+	Ref toCall = arena->external(S("complexIntFn"), address(&complexIntFn));
 
 	Listing *l = new (e) Listing(false, intDesc(e));
 	Var params = l->createParam(ptrDesc(e));
@@ -548,20 +621,6 @@ BEGIN_TEST(CallComplexFromArray, Code) {
 
 } END_TEST
 
-struct ByteStruct {
-	Byte a;
-	Byte b;
-
-	ByteStruct(Byte a, Byte b) : a(a), b(b) {}
-};
-
-SimpleDesc *bytesDesc(Engine &e) {
-	SimpleDesc *desc = new (e) SimpleDesc(Size::sByte * 2, 2);
-	desc->at(0) = Primitive(primitive::integer, Size::sByte, Offset());
-	desc->at(1) = Primitive(primitive::integer, Size::sByte, Offset::sByte);
-	return desc;
-}
-
 ByteStruct CODECALL callBytes(ByteStruct p) {
 	return ByteStruct(p.a + 10, p.b - 5);
 }
@@ -598,3 +657,72 @@ BEGIN_TEST(CallBytes, Code) {
 	CHECK_EQ(r.a, 33);
 	CHECK_EQ(r.b, 16);
 } END_TEST
+
+Float CODECALL callRect(geometry::Rect r) {
+	// On ARM64 it seems like each fp value is passed in its own vector register.
+	return r.size().w + r.size().h;
+}
+
+geometry::Rect returnRect() {
+	return geometry::Rect(1, 2, 3, 4);
+}
+
+BEGIN_TEST(CallRect, Code) {
+	Engine &e = gEngine();
+	Arena *arena = code::arena(e);
+
+	Ref rectFn = arena->external(S("rectFn"), address(&callRect));
+	SimpleDesc *rect = rectDesc(e);
+
+	Listing *l = new (e) Listing(false, floatDesc(e));
+	Var p = l->createVar(l->root(), rect->size());
+
+	*l << prolog();
+	*l << mov(floatRel(p, Offset()), floatConst(10.0f));
+	*l << mov(floatRel(p, Offset::sFloat), floatConst(11.0f));
+	*l << mov(floatRel(p, Offset::sFloat * 2), floatConst(21.0f));
+	*l << mov(floatRel(p, Offset::sFloat * 3), floatConst(31.0f));
+	*l << fnParam(rect, p);
+	*l << fnCall(rectFn, false, floatDesc(e), eax);
+	*l << fnRet(eax);
+
+	Binary *bin = new (e) Binary(arena, l);
+	typedef Float (*Fn)();
+	Fn fn = (Fn)bin->address();
+
+	Float f = (*fn)();
+	CHECK_EQ(f, 31.0f);
+} END_TEST
+
+BEGIN_TEST(CallRectRet, Code) {
+	Engine &e = gEngine();
+	Arena *arena = code::arena(e);
+
+	Ref rectRetFn = arena->external(S("rectRetFn"), address(&returnRect));
+	SimpleDesc *rect = rectDesc(e);
+
+	Listing *l = new (e) Listing(false, floatDesc(e));
+	Var p = l->createVar(l->root(), rect->size());
+
+	*l << prolog();
+	*l << fnCall(rectRetFn, false, rect, p);
+	*l << mov(eax, floatRel(p, Offset()));
+	*l << fadd(eax, floatRel(p, Offset::sFloat));
+	*l << fadd(eax, floatRel(p, Offset::sFloat*2));
+	*l << fadd(eax, floatRel(p, Offset::sFloat*3));
+	*l << fnRet(eax);
+
+	Binary *bin = new (e) Binary(arena, l);
+	typedef Float (*Fn)();
+	Fn fn = (Fn)bin->address();
+
+	Float f = (*fn)();
+	CHECK_EQ(f, 10.0f);
+} END_TEST
+
+BEGIN_TEST(CallMore, Code) {
+	TODO(L"Implement more tests!");
+	// The following tests are needed:
+	// - Function call that has complex parameters, then uses value in eax (probably crashes x64 backend as well)
+	// - Receiving simple and complex parameters that get passed in memory.
+} END_TEST
