@@ -402,6 +402,9 @@ typedef struct ScanStateStruct {
   Sig sig;                      /* <design/sig> */
   struct mps_ss_s ss_s;         /* .ss <http://bash.org/?400459> */
   Arena arena;                  /* owning arena */
+  mps_fmt_scan_t formatScan;    /* callback for scanning formatted objects */
+  mps_area_scan_t areaScan;     /* ditto via the area scanning interface */
+  void *areaScanClosure;        /* closure argument for areaScan */
   SegFixMethod fix;             /* third stage fix function */
   void *fixClosure;             /* see .ss.fix-closure */
   TraceSet traces;              /* traces to scan for */
