@@ -37,6 +37,9 @@ Bool ScanStateCheck(ScanState ss)
   ZoneSet white;
 
   CHECKS(ScanState, ss);
+  CHECKL(FUNCHECK(ss->formatScan));
+  CHECKL(FUNCHECK(ss->areaScan));
+  /* Can't check ss->areaScanClosure. */
   CHECKL(FUNCHECK(ss->fix));
   /* Can't check ss->fixClosure. */
   CHECKL(ScanStateZoneShift(ss) == ss->arena->zoneShift);
@@ -56,6 +59,15 @@ Bool ScanStateCheck(ScanState ss)
 }
 
 
+/* traceNoAreaScan -- area scan function that must not be called */
+
+static mps_res_t traceNoAreaScan(mps_ss_t ss, void *base, void *limit, void *closure)
+{
+  UNUSED(closure);
+  return FormatNoScan(ss, base, limit);
+}
+
+
 /* ScanStateInit -- Initialize a ScanState object */
 
 void ScanStateInit(ScanState ss, TraceSet ts, Arena arena,
@@ -91,6 +103,8 @@ void ScanStateInit(ScanState ss, TraceSet ts, Arena arena,
   if (ss->fix == SegFix && ArenaEmergency(arena))
         ss->fix = SegFixEmergency;
 
+  ss->formatScan = FormatNoScan;
+  ss->areaScan = traceNoAreaScan;
   ss->rank = rank;
   ss->traces = ts;
   ScanStateSetZoneShift(ss, arena->zoneShift);
@@ -114,6 +128,21 @@ void ScanStateInit(ScanState ss, TraceSet ts, Arena arena,
 }
 
 
+/* ScanStateInitSeg -- Initialize a ScanState object for scanning a segment */
+
+void ScanStateInitSeg(ScanState ss, TraceSet ts, Arena arena,
+                      Rank rank, ZoneSet white, Seg seg)
+{
+  Format format;
+  AVERT(Seg, seg);
+
+  ScanStateInit(ss, ts, arena, rank, white);
+  if (PoolFormat(&format, SegPool(seg))) {
+    ss->formatScan = format->scan;
+  }
+}
+
+
 /* ScanStateFinish -- Finish a ScanState object */
 
 void ScanStateFinish(ScanState ss)
@@ -1114,6 +1143,34 @@ RefSet ScanStateSummary(ScanState ss)
 }
 
 
+/* ScanStateUpdateSummary -- update segment summary after scan
+ *
+ * wasTotal is TRUE if we know that all references were scanned, FALSE
+ * if some references might not have been scanned.
+ */
+void ScanStateUpdateSummary(ScanState ss, Seg seg, Bool wasTotal)
+{
+  RefSet summary;
+
+  AVERT(ScanState, ss);
+  AVERT(Seg, seg);
+  AVERT(Bool, wasTotal);
+
+  /* Only apply the write barrier if it is not deferred. */
+  if (seg->defer == 0) {
+    /* If we scanned every reference in the segment then we have a
+       complete summary we can set. Otherwise, we just have
+       information about more zones that the segment refers to. */
+    if (wasTotal)
+      summary = ScanStateSummary(ss);
+    else
+      summary = RefSetUnion(SegSummary(seg), ScanStateSummary(ss));
+  } else {
+    summary = RefSetUNIV;
+  }
+  SegSetSummary(seg, summary);
+}
+
 /* traceScanSegRes -- scan a segment to remove greyness
  *
  * @@@@ During scanning, the segment should be write-shielded to prevent
@@ -1126,7 +1183,6 @@ static Res traceScanSegRes(TraceSet ts, Rank rank, Arena arena, Seg seg)
   Bool wasTotal;
   ZoneSet white;
   Res res;
-  RefSet summary;
 
   /* The reason for scanning a segment is that it's grey. */
   AVER(TraceSetInter(ts, SegGrey(seg)) != TraceSetEMPTY);
@@ -1141,7 +1197,7 @@ static Res traceScanSegRes(TraceSet ts, Rank rank, Arena arena, Seg seg)
   } else {      /* scan it */
     ScanStateStruct ssStruct;
     ScanState ss = &ssStruct;
-    ScanStateInit(ss, ts, arena, rank, white);
+    ScanStateInitSeg(ss, ts, arena, rank, white, seg);
 
     /* Expose the segment to make sure we can scan it. */
     ShieldExpose(arena, seg);
@@ -1183,20 +1239,7 @@ static Res traceScanSegRes(TraceSet ts, Rank rank, Arena arena, Seg seg)
         seg->defer = WB_DEFER_DELAY;
     }
 
-    /* Only apply the write barrier if it is not deferred. */
-    if (seg->defer == 0) {
-      /* If we scanned every reference in the segment then we have a
-         complete summary we can set. Otherwise, we just have
-         information about more zones that the segment refers to. */
-      if (res == ResOK && wasTotal)
-        summary = ScanStateSummary(ss);
-      else
-        summary = RefSetUnion(SegSummary(seg), ScanStateSummary(ss));
-    } else {
-      summary = RefSetUNIV;
-    }
-    SegSetSummary(seg, summary);
-
+    ScanStateUpdateSummary(ss, seg, res == ResOK && wasTotal);
     ScanStateFinish(ss);
   }
 
@@ -1473,16 +1516,36 @@ void TraceScanSingleRef(TraceSet ts, Rank rank, Arena arena,
 }
 
 
+/* TraceScanFormat -- scan a formatted area of memory for references
+ *
+ * This is a wrapper for format scanning functions, which should not
+ * otherwise be called directly from within the MPS.  This function
+ * checks arguments and takes care of accounting for the scanned
+ * memory.
+ */
+Res TraceScanFormat(ScanState ss, Addr base, Addr limit)
+{
+  AVERT_CRITICAL(ScanState, ss);
+  AVER_CRITICAL(base != NULL);
+  AVER_CRITICAL(limit != NULL);
+  AVER_CRITICAL(base < limit);
+
+  /* scannedSize is accumulated whether or not ss->formatScan
+   * succeeds, so it's safe to accumulate now so that we can tail-call
+   * ss->formatScan. */
+  ss->scannedSize += AddrOffset(base, limit);
+
+  return ss->formatScan(&ss->ss_s, base, limit);
+}
+
+
 /* TraceScanArea -- scan an area of memory for references
  *
  * This is a wrapper for area scanning functions, which should not
  * otherwise be called directly from within the MPS.  This function
  * checks arguments and takes care of accounting for the scanned
  * memory.
- *
- * c.f. FormatScan()
  */
-
 Res TraceScanArea(ScanState ss, Word *base, Word *limit,
                   mps_area_scan_t scan_area,
                   void *closure)
